# SPDX-License-Identifier: Apache-2.0

import inspect
import unittest
import os
import numpy
import pandas
from pyspark.ml import PipelineModel
from onnxmltools import convert_sparkml
from onnxmltools.convert.sparkml import buildInitialTypesSimple, buildInputDictSimple
from onnxmltools.utils.utils_backend_onnxruntime import (
    run_with_runtime,
    _compare_expected,
)
from tests.sparkml import SparkMlTestCase


class RPipeline(SparkMlTestCase):
    def test_sparkml_r_pipeline(self):
        # add additional jar files before creating SparkSession
        this_script_dir = os.path.dirname(
            os.path.abspath(inspect.getfile(inspect.currentframe()))
        )
        input_path = os.path.join(this_script_dir, "data", "iris.csv")
        data = (
            self.spark.read.format("csv")
            .options(header="true", inferschema="true")
            .load(input_path)
            .drop("_index_")
        )

        # read the model from disk
        pipeline_path = os.path.join(this_script_dir, "mlpmodel")
        model = PipelineModel.load(path=pipeline_path)

        # create Onnx model
        model_onnx = convert_sparkml(
            model,
            "Sparkml R Pipeline",
            buildInitialTypesSimple(data),
            spark_session=self.spark,
        )
        # save Onnx model for runtime usage
        if model_onnx is None:
            raise AssertionError("Failed to create the onnx model")
        model_path = os.path.join(
            this_script_dir, "tests_dump", "r_pipeline_model.onnx"
        )
        with open(model_path, "wb") as f:
            f.write(model_onnx.SerializeToString())

        data_np = buildInputDictSimple(data)
        # run the model in Spark
        spark_prediction = model.transform(data)
        # run the model in onnx runtime
        output, session = run_with_runtime(data_np, model_path)

        # compare results
        expected = [
            spark_prediction.toPandas().label.values.astype(numpy.float32),
            spark_prediction.toPandas().prediction.values.astype(numpy.float32),
            spark_prediction.toPandas()
            .probability.apply(lambda x: pandas.Series(x.toArray()))
            .values.astype(numpy.float32),
        ]
        _compare_expected(
            expected, output, session, model_path, decimal=5, onnx_shape=None
        )


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_shape_calculator
from ...common.data_types import DictionaryType, SequenceType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculate_dictionary_vectorizer_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. Map ---> [1, C]

    C is the number of all allowed keys in the input dictionary.
    """
    # We assume all dictionaries' value types are float. It seems be reasonable to CoreML's
    # model input, but the existence of other map types leads to some concerns.
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    # Two types are allowed. One is DictionaryType and
    # the other one, SequenceType, means a sequence of dictionaries.
    check_input_and_output_types(
        operator, good_input_types=[DictionaryType, SequenceType]
    )

    params = operator.raw_operator.dictVectorizer
    string_key_vector = params.stringToIndex.vector
    int64_key_vector = params.int64ToIndex.vector

    if len(string_key_vector) > 0 and len(int64_key_vector) > 0:
        raise RuntimeError("Only one key type can present at the same time")

    doc_string = operator.outputs[0].type.doc_string
    if len(string_key_vector) > 0:
        operator.outputs[0].type = FloatTensorType(
            [1, len(string_key_vector)], doc_string=doc_string
        )
    elif len(int64_key_vector) > 0:
        operator.outputs[1].type.shape = FloatTensorType(
            [1, len(int64_key_vector)], doc_string=doc_string
        )
    else:
        raise ValueError("Key vector cannot be empty")


register_shape_calculator(
    "dictVectorizer", calculate_dictionary_vectorizer_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

from onnx import onnx_pb as onnx_proto
from pyspark.ml.classification import GBTClassificationModel

from ...common._apply_operation import (
    apply_neg,
    apply_concat,
    apply_mul,
    apply_exp,
    apply_add,
    apply_argmax,
    apply_matmul,
)
from ...common.data_types import Int64TensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._registration import register_converter, register_shape_calculator
from ..ops_names import get_sparkml_operator_name
from .decision_tree_regressor import convert_decision_tree_regressor


def convert_gbt_classifier(scope, operator, container):
    op = operator.raw_operator
    regressor_output_names = []
    # spark implementation uses DecisionTreeRegressor
    # (and not Classifier) for each tree in this forest
    for tree_model in op.trees:
        regressor_op = scope.declare_local_operator(
            get_sparkml_operator_name(type(tree_model)), tree_model
        )
        regressor_op.raw_params = operator.raw_params
        regressor_op.inputs = operator.inputs
        regressor_output = scope.declare_local_variable(
            "regressor_prediction", FloatTensorType()
        )
        regressor_output_names.append(regressor_output.full_name)
        regressor_op.outputs.append(regressor_output)
        convert_decision_tree_regressor(scope, regressor_op, container)
        regressor_op.is_evaluated = True

    targets_tensor = scope.get_unique_variable_name("target_tensor")
    weights_tensor = scope.get_unique_variable_name("weights_tensor")
    container.add_initializer(
        weights_tensor,
        onnx_proto.TensorProto.FLOAT,
        [len(op.treeWeights), 1],
        op.treeWeights,
    )
    concatenated_predictions = scope.get_unique_variable_name(
        "concatenated_predictions_tensor"
    )
    apply_concat(
        scope, regressor_output_names, concatenated_predictions, container, axis=1
    )
    apply_matmul(
        scope, [concatenated_predictions, weights_tensor], targets_tensor, container
    )

    # this is to calculate prediction and probability
    # given the raw_prediction (= [-target, target])
    targets_neg_tensor = scope.get_unique_variable_name("target_neg_tensor")
    apply_neg(scope, targets_tensor, targets_neg_tensor, container)
    raw_prediction_tensor = scope.get_unique_variable_name("raw_prediction_tensor")
    apply_concat(
        scope,
        [targets_neg_tensor, targets_tensor],
        raw_prediction_tensor,
        container,
        axis=1,
    )
    if isinstance(op, GBTClassificationModel):
        # this section is only for the classifier;
        # for the regressor we don't calculate the probability
        minus_two = scope.get_unique_variable_name("minus_two_tensor")
        container.add_initializer(minus_two, onnx_proto.TensorProto.FLOAT, [1], [-2.0])
        mul_output_tensor = scope.get_unique_variable_name("mul_output_tensor")
        apply_mul(
            scope, [raw_prediction_tensor, minus_two], mul_output_tensor, container
        )
        exp_output_tensor = scope.get_unique_variable_name("exp_output_tensor")
        apply_exp(scope, mul_output_tensor, exp_output_tensor, container)
        one_tensor = scope.get_unique_variable_name("one_tensor")
        container.add_initializer(one_tensor, onnx_proto.TensorProto.FLOAT, [1], [1.0])
        add_output_tensor = scope.get_unique_variable_name("add_output_tensor")
        apply_add(scope, [exp_output_tensor, one_tensor], add_output_tensor, container)
        container.add_node(
            "Reciprocal",
            add_output_tensor,
            operator.outputs[1].full_name,
            name=scope.get_unique_operator_name("Reciprocal"),
            op_version=6,
        )
    # to get Prediction from rawPrediction (or probability)
    apply_argmax(
        scope,
        raw_prediction_tensor,
        operator.outputs[0].full_name,
        container,
        axis=1,
        keepdims=0,
    )


register_converter(
    "pyspark.ml.classification.GBTClassificationModel", convert_gbt_classifier
)
register_converter("pyspark.ml.regression.GBTRegressionModel", convert_gbt_classifier)


def calculate_gbt_classifier_output_shapes(operator):
    check_input_and_output_numbers(
        operator, input_count_range=1, output_count_range=[1, 2]
    )
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType]
    )
    if len(operator.inputs[0].type.shape) != 2:
        raise RuntimeError("Input must be a [N, C]-tensor")

    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = Int64TensorType(shape=[N])
    if isinstance(operator.raw_operator, GBTClassificationModel):
        class_count = 2
        operator.outputs[1].type = FloatTensorType([N, class_count])


register_shape_calculator(
    "pyspark.ml.classification.GBTClassificationModel",
    calculate_gbt_classifier_output_shapes,
)
register_shape_calculator(
    "pyspark.ml.regression.GBTRegressionModel", calculate_gbt_classifier_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

import tempfile
import os
import time
import numpy
import re
from pyspark.sql import SparkSession


class SparkMLTree(dict):
    pass


def sparkml_tree_dataset_to_sklearn(tree_df, is_classifier):
    feature = []
    threshold = []
    tree_pandas = tree_df.toPandas().sort_values("id")
    children_left = tree_pandas.leftChild.values.tolist()
    children_right = tree_pandas.rightChild.values.tolist()
    ids = tree_pandas.id.values.tolist()
    if is_classifier:
        value = numpy.array(tree_pandas.impurityStats.values.tolist())
    else:
        value = tree_pandas.prediction.values.tolist()

    for item in tree_pandas.split:
        if isinstance(item, dict):
            try:
                feature.append(item["featureIndex"])
                threshold.append(item["leftCategoriesOrThreshold"])
            except KeyError:
                raise RuntimeError(f"Unable to process {item}.")
        else:
            tuple(item)
            feature.append(item[0])
            threshold.append(item[1][0] if len(item[1]) >= 1 else -1.0)

    tree = SparkMLTree()
    tree.nodes_ids = ids
    tree.children_left = children_left
    tree.children_right = children_right
    tree.value = numpy.asarray(value, dtype=numpy.float32)
    tree.feature = feature
    tree.threshold = threshold
    tree.node_count = tree_df.count()
    return tree


def save_read_sparkml_model_data(spark: SparkSession, model):
    # Get the value of spark.master
    spark_mode = spark.conf.get("spark.master")

    # Check the value of spark.master using regular expression
    if "spark://" in spark_mode and (
        "localhost" not in spark_mode or "127.0.0.1" not in spark_mode
    ):
        dfs_key = "ONNX_DFS_PATH"
        try:
            dfs_path = spark.conf.get("ONNX_DFS_PATH")
        except Exception:
            raise ValueError(
                "Configuration property '{}' does not exist for SparkSession. \
                Please set this variable to a root distributed file system path to allow \
                for saving and reading of spark models in cluster mode. \
                You can set this in your SparkConfig \
                by setting sparkBuilder.config(ONNX_DFS_PATH, dfs_path)".format(
                    dfs_key
                )
            )
        if dfs_path is None:
            # If dfs_path is not specified, throw an error message
            # dfs_path arg is required for cluster mode
            raise ValueError(
                "Argument dfs_path is required for saving model '{}' in cluster mode. \
                You can set this in your SparkConfig by \
                setting sparkBuilder.config(ONNX_DFS_PATH, dfs_path)".format(
                    type(model).__name__
                )
            )
        else:
            # Check that the dfs_path is a valid distributed file system path
            # This can be hdfs, wabs, s3, etc.
            if re.match(r"^[a-zA-Z]+://", dfs_path) is None:
                raise ValueError(
                    "Argument dfs_path '{}' is not a valid distributed path".format(
                        dfs_path
                    )
                )
            else:
                # If dfs_path is specified, save the model to a tmp directory
                # The dfs_path will be the root of the /tmp
                tdir = os.path.join(dfs_path, "tmp/onnx")
    else:
        # If spark.master is not set or set to local, save the model to a local path.
        tdir = tempfile.tempdir
        if tdir is None:
            local_dir = spark._jvm.org.apache.spark.util.Utils.getLocalDir(
                spark._jsc.sc().conf()
            )
            tdir = spark._jvm.org.apache.spark.util.Utils.createTempDir(
                local_dir, "onnx"
            ).getAbsolutePath()
        if tdir is None:
            raise FileNotFoundError(
                "Unable to create a temporary directory for model '{}'"
                ".".format(type(model).__name__)
            )

    path = os.path.join(tdir, type(model).__name__ + "_" + str(time.time()))
    model.write().overwrite().save(path)
    df = spark.read.parquet(os.path.join(path, "data"))
    return df


def get_default_tree_classifier_attribute_pairs():
    attrs = {}
    attrs["post_transform"] = "NONE"
    attrs["nodes_treeids"] = []
    attrs["nodes_nodeids"] = []
    attrs["nodes_featureids"] = []
    attrs["nodes_modes"] = []
    attrs["nodes_values"] = []
    attrs["nodes_truenodeids"] = []
    attrs["nodes_falsenodeids"] = []
    attrs["nodes_missing_value_tracks_true"] = []
    attrs["nodes_hitrates"] = []
    attrs["class_treeids"] = []
    attrs["class_nodeids"] = []
    attrs["class_ids"] = []
    attrs["class_weights"] = []
    return attrs


def get_default_tree_regressor_attribute_pairs():
    attrs = {}
    attrs["post_transform"] = "NONE"
    attrs["n_targets"] = 0
    attrs["nodes_treeids"] = []
    attrs["nodes_nodeids"] = []
    attrs["nodes_featureids"] = []
    attrs["nodes_modes"] = []
    attrs["nodes_values"] = []
    attrs["nodes_truenodeids"] = []
    attrs["nodes_falsenodeids"] = []
    attrs["nodes_missing_value_tracks_true"] = []
    attrs["nodes_hitrates"] = []
    attrs["target_treeids"] = []
    attrs["target_nodeids"] = []
    attrs["target_ids"] = []
    attrs["target_weights"] = []
    return attrs


def add_node(
    attr_pairs,
    is_classifier,
    tree_id,
    tree_weight,
    node_id,
    feature_id,
    mode,
    value,
    true_child_id,
    false_child_id,
    weights,
    weight_id_bias,
    leaf_weights_are_counts,
):
    attr_pairs["nodes_treeids"].append(tree_id)
    attr_pairs["nodes_nodeids"].append(node_id)
    attr_pairs["nodes_featureids"].append(feature_id)
    attr_pairs["nodes_modes"].append(mode)
    attr_pairs["nodes_values"].append(value)
    attr_pairs["nodes_truenodeids"].append(true_child_id)
    attr_pairs["nodes_falsenodeids"].append(false_child_id)
    attr_pairs["nodes_missing_value_tracks_true"].append(False)
    attr_pairs["nodes_hitrates"].append(1.0)

    # Add leaf information for making prediction
    if mode == "LEAF":
        flattened_weights = weights.flatten()
        factor = tree_weight
        # If the values stored at leaves are counts of possible classes,
        # we need convert them to probabilities by
        # doing a normalization.
        if leaf_weights_are_counts:
            s = sum(flattened_weights)
            factor /= float(s) if s != 0.0 else 1.0
        flattened_weights = [w * factor for w in flattened_weights]
        if len(flattened_weights) == 2 and is_classifier:
            flattened_weights = [flattened_weights[1]]

        # Note that attribute names for making prediction
        # are different for classifiers and regressors
        if is_classifier:
            for i, w in enumerate(flattened_weights):
                attr_pairs["class_treeids"].append(tree_id)
                attr_pairs["class_nodeids"].append(node_id)
                attr_pairs["class_ids"].append(i + weight_id_bias)
                attr_pairs["class_weights"].append(w)
        else:
            for i, w in enumerate(flattened_weights):
                attr_pairs["target_treeids"].append(tree_id)
                attr_pairs["target_nodeids"].append(node_id)
                attr_pairs["target_ids"].append(i + weight_id_bias)
                attr_pairs["target_weights"].append(w)


def add_tree_to_attribute_pairs(
    attr_pairs,
    is_classifier,
    tree,
    tree_id,
    tree_weight,
    weight_id_bias,
    leaf_weights_are_counts,
):
    for i in range(tree.node_count):
        node_id = tree.nodes_ids[i]
        weight = tree.value[i]

        if tree.children_left[i] >= 0 or tree.children_right[i] >= 0:
            mode = "BRANCH_LEQ"
            feat_id = tree.feature[i]
            threshold = tree.threshold[i]
            left_child_id = int(tree.children_left[i])
            right_child_id = int(tree.children_right[i])
        else:
            mode = "LEAF"
            feat_id = 0
            threshold = 0.0
            left_child_id = 0
            right_child_id = 0

        add_node(
            attr_pairs,
            is_classifier,
            tree_id,
            tree_weight,
            node_id,
            feat_id,
            mode,
            threshold,
            left_child_id,
            right_child_id,
            weight,
            weight_id_bias,
            leaf_weights_are_counts,
        )



# SPDX-License-Identifier: Apache-2.0

from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import (
    get_maximum_opset_supported as _get_maximum_opset_supported,
)


def get_maximum_opset_supported():
    return min(onnx_opset_version(), max(15, _get_maximum_opset_supported()))



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter


def convert_one_hot_encoder(scope, operator, container):
    op_type = "OneHotEncoder"
    attrs = {"name": operator.full_name}

    raw_model = operator.raw_operator.oneHotEncoder
    if raw_model.HasField("int64Categories"):
        attrs["cats_int64s"] = list(int(i) for i in raw_model.int64Categories.vector)
    if raw_model.HasField("stringCategories"):
        attrs["cats_strings"] = list(
            str(s).encode("utf-8") for s in raw_model.stringCategories.vector
        )

    container.add_node(
        op_type,
        [operator.inputs[0].full_name],
        [operator.outputs[0].full_name],
        op_domain="ai.onnx.ml",
        **attrs
    )


register_converter("oneHotEncoder", convert_one_hot_encoder)



# SPDX-License-Identifier: Apache-2.0

from ...common.data_types import (
    Int64TensorType,
    Int64Type,
    FloatTensorType,
    FloatType,
    StringType,
)


def convert_integer_to_float(scope, variable, container):
    op_type = "Scaler"
    scaled_name = scope.get_unique_variable_name(variable.full_name + "_scaled")
    scaler_attrs = {
        "name": scope.get_unique_operator_name(op_type),
        "scale": [1.0],
        "offset": [0.0],
    }
    container.add_node(
        "Scaler",
        variable.full_name,
        scaled_name,
        op_domain="ai.onnx.ml",
        **scaler_attrs
    )
    return scaled_name


def concatenate_variables(scope, variables, container):
    """
    This function allocate operators to from a float
    tensor by concatenating all input variables. Notice that if all
    integer inputs would be converted to floats before concatenation.
    """

    # Check if it's possible to concatenate those inputs.
    type_set = set(type(variable.type) for variable in variables)
    number_type_set = {FloatType, FloatTensorType, Int64Type, Int64TensorType}
    if StringType in type_set and any(
        number_type in type_set for number_type in number_type_set
    ):
        raise RuntimeError(
            "We are not able to concatenate numerical tensor(s) and string tensor(s)"
        )

    input_names = []  # input variables' names we want to concatenate
    input_dims = []  # dimensions of the variables that is going to be concatenated

    # Collect input variable names and do cast if needed
    for variable in variables:
        if isinstance(variable.type, (Int64TensorType, Int64Type)):
            input_names.append(convert_integer_to_float(scope, variable, container))
        else:
            input_names.append(variable.full_name)
        # We assume input variables' shape are [1, C_1], ..., [1, C_n] if there are n inputs.
        input_dims.append(variable.type.shape[1])

    if len(input_names) == 1:
        # No need to concatenate tensors if there is only one input
        return input_names[0]
    else:
        # To combine all inputs, we need a FeatureVectorizer
        op_type = "FeatureVectorizer"
        attrs = {
            "name": scope.get_unique_operator_name(op_type),
            "inputdimensions": input_dims,
        }
        # Create a variable name to capture feature vectorizer's output
        concatenated_name = scope.get_unique_variable_name("concatenated")
        # Set up our FeatureVectorizer
        container.add_node(
            op_type, input_names, concatenated_name, op_domain="ai.onnx.ml", **attrs
        )

        return concatenated_name



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import packaging.version as pv
import onnx
import pandas
import numpy

try:
    from sklearn.utils._testing import ignore_warnings
except ImportError:
    from sklearn.utils.testing import ignore_warnings
from pyspark.ml.classification import DecisionTreeClassifier
from pyspark.ml.feature import VectorAssembler
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import (
    save_data_models,
    compare_results,
    run_onnx_model,
)
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmDecisionTreeClassifierBig(SparkMlTestCase):
    # @unittest.skipIf(True, reason="Mismatched input dimensions.")
    @ignore_warnings(category=(ResourceWarning, DeprecationWarning))
    @unittest.skipIf(sys.platform == "win32", reason="UnsatisfiedLinkError")
    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    @unittest.skipIf(
        pv.Version(onnx.__version__) <= pv.Version("1.3"), "Need Greater Opset 9"
    )
    def test_tree_pipeline_category(self):
        ok = {1, 6, 10, 56, 78, 34}
        ok2 = {1, 7, 8, 23, 35, 45}

        def f_label(x):
            if x[1] in ok or x[0] in ok:
                return 1
            if x[1] in ok or x[0] in ok2:
                return 2
            return 0

        features = numpy.random.randint(0, high=50, size=(1000, 2)) % 100
        labels = numpy.array([f_label(row) for row in features])
        n_features = features.shape[1]
        df = pandas.DataFrame(features)
        features_names = [f"c{i}" for i in range(df.shape[1])]
        df.columns = features_names
        cat = set(s for s in df["c0"])
        df["c0"] = pandas.Categorical(
            list(map(lambda s: int(s), df["c0"])), categories=cat, ordered=False
        )
        cat = set(s for s in df["c1"])
        df["c1"] = pandas.Categorical(
            list(map(lambda s: int(s), df["c1"])), categories=cat, ordered=False
        )
        df["label"] = labels

        sparkDF = self.spark.createDataFrame(df)

        data = sparkDF  # self.spark.read.csv(input_path, header=True, inferSchema=True)
        va = VectorAssembler(inputCols=features_names, outputCol="features")
        va_df = va.transform(data)
        va_df = va_df.select(["features", "label"])

        dt = DecisionTreeClassifier(
            labelCol="label", featuresCol="features", maxDepth=3, maxBins=50
        )
        model = dt.fit(va_df)
        # print(model.toDebugString)
        model_onnx = convert_sparkml(
            model,
            "Sparkml Decision Tree Binary Class",
            [("features", FloatTensorType([None, n_features]))],
            spark_session=self.spark,
            target_opset=TARGET_OPSET,
        )
        data_np = (
            va_df.toPandas()
            .features.apply(lambda x: pandas.Series(x.toArray()))
            .values.astype(numpy.float32)
        )
        predicted = model.transform(va_df)
        expected = [
            predicted.toPandas().prediction.values.astype(numpy.float32),
            predicted.toPandas()
            .probability.apply(lambda x: pandas.Series(x.toArray()))
            .values.astype(numpy.float32),
        ]
        paths = save_data_models(
            data_np,
            expected,
            model,
            model_onnx,
            basename="SparkmlDecisionTreeBinaryClassCategory",
        )
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(
            ["prediction", "probability"], data_np, onnx_model_path
        )
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    import logging

    logging.basicConfig(encoding="utf-8", level=logging.INFO)
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

from .....proto import helper
from .....proto import onnx_proto
from ....common._registration import register_converter
from ....common._apply_operation import apply_constant


def convert_load_constant_nd(scope, operator, container):
    params = operator.raw_operator.loadConstantND
    constant_name = scope.get_unique_variable_name("constant")
    constant = helper.make_tensor(
        constant_name,
        onnx_proto.TensorProto.FLOAT,
        params.shape,
        params.data.floatValue,
    )

    apply_constant(
        scope,
        operator.output_full_names,
        container,
        operator_name=operator.full_name,
        value=constant,
    )


register_converter("loadConstantND", convert_load_constant_nd)



# SPDX-License-Identifier: Apache-2.0

import copy
from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import (
    check_input_and_output_numbers,
    check_input_and_output_types,
)


def calculate_crop_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C, H, W] ---> [N, C, H', W']
        2. [N, C, H, W],  shape-ref [N', C', H', W'] ---> [N, C, H', W']
    """
    check_input_and_output_numbers(
        operator, input_count_range=[1, 2], output_count_range=1
    )
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    output_shape = copy.deepcopy(operator.inputs[0].type.shape)

    params = operator.raw_operator.crop
    if len(operator.inputs) == 1:
        if len(params.cropAmounts.borderAmounts) > 0:
            output_shape[2] -= params.cropAmounts.borderAmounts[0].startEdgeSize
            output_shape[2] -= params.cropAmounts.borderAmounts[0].endEdgeSize
            output_shape[3] -= params.cropAmounts.borderAmounts[1].startEdgeSize
            output_shape[3] -= params.cropAmounts.borderAmounts[1].endEdgeSize
    elif len(operator.inputs) == 2:
        output_shape[2] = operator.inputs[1].type.shape[2]
        output_shape[3] = operator.inputs[1].type.shape[3]
    else:
        raise RuntimeError("Too many inputs for Crop operator")

    operator.outputs[0].type.shape = output_shape


register_shape_calculator("crop", calculate_crop_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_shape_calculator
from ...common.data_types import (
    FloatTensorType,
    Int64TensorType,
    FloatType,
    Int64Type,
    DictionaryType,
    SequenceType,
    StringTensorType,
)
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculate_traditional_classifier_output_shapes(operator):
    """
    For classifiers, allowed input/output patterns are
        1. [N, C_1], ..., [N, C_n] ---> [N], Sequence of Map
        2. [N, C_1], ..., [N, C_n] ---> [N]

    For regressors, allowed input/output patterns are
        1. [N, C_1], ..., [N, C_n] ---> [N, 1]

    Core ML classifiers and regressors support multiple
    input feature tensors, so we need to concatenate them before
    feeding them into their ONNX counterparts.
    Note that the N must be 1 as long as ZipMap only produces dictionary.
    """
    check_input_and_output_numbers(
        operator, input_count_range=[1, None], output_count_range=[1, 2]
    )
    check_input_and_output_types(
        operator,
        good_input_types=[FloatTensorType, Int64TensorType, FloatType, Int64Type],
    )

    if any(len(variable.type.shape) != 2 for variable in operator.inputs):
        raise RuntimeError("Input(s) must be [N, C]-tensor(s)")

    model_type = operator.raw_operator.WhichOneof("Type")
    if model_type == "treeEnsembleClassifier":
        class_label_type = operator.raw_operator.treeEnsembleClassifier.WhichOneof(
            "ClassLabels"
        )
    elif model_type == "glmClassifier":
        class_label_type = operator.raw_operator.glmClassifier.WhichOneof("ClassLabels")
    elif model_type == "supportVectorClassifier":
        class_label_type = operator.raw_operator.supportVectorClassifier.WhichOneof(
            "ClassLabels"
        )
    else:
        raise ValueError("%s has no class label" % model_type)

    N = operator.inputs[0].type.shape[0]
    if operator.target_opset < 7:
        output_shape = [1, 1]
    else:
        output_shape = [N]

    if class_label_type == "stringClassLabels":
        operator.outputs[0].type = StringTensorType(
            output_shape, doc_string=operator.outputs[0].type.doc_string
        )
        if len(operator.outputs) == 2:
            if operator.target_opset < 7:
                operator.outputs[1].type = DictionaryType(
                    StringTensorType([1]),
                    FloatTensorType([1]),
                    doc_string=operator.outputs[1].type.doc_string,
                )
            else:
                operator.outputs[1].type = SequenceType(
                    DictionaryType(StringTensorType([]), FloatTensorType([])),
                    doc_string=operator.outputs[1].type.doc_string,
                )
    elif class_label_type == "int64ClassLabels":
        operator.outputs[0].type = Int64TensorType(
            output_shape, doc_string=operator.outputs[0].type.doc_string
        )
        if len(operator.outputs) == 2:
            if operator.target_opset < 7:
                operator.outputs[1].type = DictionaryType(
                    Int64TensorType([1]),
                    FloatTensorType([1]),
                    doc_string=operator.outputs[1].type.doc_string,
                )
            else:
                operator.outputs[1].type = SequenceType(
                    DictionaryType(Int64TensorType([]), FloatTensorType([])),
                    doc_string=operator.outputs[1].type.doc_string,
                )
    else:
        raise ValueError("Traditional classifier must include label information")


register_shape_calculator(
    "glmClassifier", calculate_traditional_classifier_output_shapes
)
register_shape_calculator(
    "supportVectorClassifier", calculate_traditional_classifier_output_shapes
)
register_shape_calculator(
    "treeEnsembleClassifier", calculate_traditional_classifier_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

import math
import numbers
from ....common._registration import register_shape_calculator
from ....common.utils import check_input_and_output_numbers


def calculate_convolution_and_pooling_1D_output_shape(
    input_size,
    kernel_size,
    kernel_dilation,
    stride,
    pad_mode,
    pad_head,
    pad_tail,
    output_size=0,
):
    if not isinstance(input_size, numbers.Integral):
        return "None"
    if output_size > 0:
        return int(output_size)  # Must use output_size = 1 for global pooling

    effective_kernel_size = 1 + kernel_dilation * (
        kernel_size - 1
    )  # For pooling, we always have dilation = 1.
    if pad_mode == "valid":
        return int(
            math.floor(
                (input_size + pad_head + pad_tail - effective_kernel_size) / stride
            )
            + 1
        )
    elif pad_mode == "same":
        return int(math.ceil(input_size / stride))
    elif pad_mode == "includeLastPixel":
        if pad_head != pad_tail:
            raise ValueError(
                "Padding amounts at the beginning and the end of an axis must be the same"
            )
        effective_input_size = input_size + pad_head + pad_tail - effective_kernel_size
        out_size = math.ceil(effective_input_size / stride) + 1
        if (out_size - 1) * stride >= input_size + pad_head:
            out_size -= 1
        return out_size
    else:
        raise ValueError("Unknown padding mode: %s" % pad_mode)


def calculate_convolution_transpose_1D_output_shape(
    input_size,
    kernel_size,
    kernel_dilation,
    stride,
    pad_mode,
    pad_head,
    pad_tail,
    output_size=0,
):
    if not isinstance(input_size, numbers.Integral):
        return "None"
    if output_size > 0:
        return output_size

    effective_kernel_size = 1 + kernel_dilation * (kernel_size - 1)
    if pad_mode == "valid":
        return int(
            (input_size - 1) * stride - pad_head - pad_tail + effective_kernel_size
        )
    elif pad_mode == "same":
        return int(input_size * stride)
    else:
        raise ValueError("Unknown padding mode: %s" % pad_mode)


def calculate_convolution_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C, H, W] ---> [N, C, H', W']
    """
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)

    params = operator.raw_operator.convolution

    input_shape = operator.inputs[0].type.shape
    operator.outputs[0].type.shape = [
        0,
        0,
        0,
        0,
    ]  # Initialize output shape. It will be modified below.
    output_shape = operator.outputs[0].type.shape

    # Adjust N-axis
    output_shape[0] = input_shape[0]

    # Adjust C-axis
    output_shape[1] = params.outputChannels

    # Set up default and non-default parameters
    dilations = [1, 1]
    if len(params.dilationFactor) > 0:
        dilations = [params.dilationFactor[0], params.dilationFactor[1]]
    kernel_shape = [3, 3]
    if len(params.kernelSize) > 0:
        kernel_shape = params.kernelSize
    strides = [1, 1]
    if len(params.stride) > 0:
        strides = params.stride
    specified_output_shape = [0, 0]  # Only used with convolution transpose
    if params.isDeconvolution and len(params.outputShape) > 0:
        specified_output_shape = list(int(i) for i in params.outputShape)
    pad_mode = params.WhichOneof("ConvolutionPaddingType")
    if pad_mode == "valid" and len(params.valid.paddingAmounts.borderAmounts) > 0:
        pad_amounts = params.valid.paddingAmounts.borderAmounts
        pad_heads = [pad_amounts[0].startEdgeSize, pad_amounts[1].startEdgeSize]
        pad_tails = [pad_amounts[0].endEdgeSize, pad_amounts[1].endEdgeSize]
    else:
        # Padding amounts are useless for same
        # padding and valid padding uses [0, 0] by default.
        pad_heads = [0, 0]
        pad_tails = [0, 0]

    # Adjust H- and W-axes
    for i in range(2):
        if params.isDeconvolution:
            output_shape[i + 2] = calculate_convolution_transpose_1D_output_shape(
                input_shape[i + 2],
                kernel_shape[i],
                dilations[i],
                strides[i],
                pad_mode,
                pad_heads[i],
                pad_tails[i],
                specified_output_shape[i],
            )
        else:
            output_shape[i + 2] = calculate_convolution_and_pooling_1D_output_shape(
                input_shape[i + 2],
                kernel_shape[i],
                dilations[i],
                strides[i],
                pad_mode,
                pad_heads[i],
                pad_tails[i],
            )


register_shape_calculator("convolution", calculate_convolution_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import itertools
import numpy as np
from ...common._registration import register_converter


def extract_support_vectors_as_dense_tensor(svm_model):
    support_type = svm_model.WhichOneof("supportVectors")
    if support_type == "denseSupportVectors":
        vectors = svm_model.denseSupportVectors.vectors
        support_vectors = np.array([v.values for v in vectors]).flatten()
    elif support_type == "sparseSupportVectors":
        # Since current ONNX doesn't support sparse representations, we have to save them as dense tensors. It may
        # dramatically prolong prediction time and increase memory usage.
        vectors = svm_model.sparseSupportVectors.vectors
        # Search for the maximum dimension over all sparse support vectors
        max_idx = 0
        for v in vectors:
            for n in v.nodes:
                if n.index > max_idx:
                    max_idx = n.index
        # Save sparse vectors to dense vectors
        support_vectors = np.zeros(shape=(len(vectors), max_idx))
        for i, v in enumerate(vectors):
            for n in v.nodes:
                support_vectors[i][n.index - 1] = n.value
        support_vectors = support_vectors.flatten()
    else:
        raise ValueError("Unsupported support vector type: %s" % support_type)
    return len(vectors), support_vectors


def convert_svm_classifier(scope, operator, container):
    params = operator.raw_operator.supportVectorClassifier
    kernel_enum = {
        "linearKernel": "LINEAR",
        "polyKernel": "POLY",
        "rbfKernel": "RBF",
        "sigmoidKernel": "SIGMOID",
        "precomputedKernel": "PRECOMPUTED",
    }
    kernel = params.kernel
    kernel_val = kernel.WhichOneof("kernel")
    svc_kernel = kernel_enum[kernel_val]

    if kernel_val == "rbfKernel":
        svc_kernel_params = [kernel.rbfKernel.gamma, 0.0, 0.0]
    elif kernel_val == "polyKernel":
        svc_kernel_params = [
            kernel.polyKernel.gamma,
            kernel.polyKernel.coef0,
            kernel.polyKernel.degree,
        ]
    elif kernel_val == "sigmoidKernel":
        svc_kernel_params = [
            kernel.sigmoidKernel.gamma,
            kernel.sigmoidKernel.coef0,
            0.0,
        ]
    elif kernel_val == "linearKernel":
        svc_kernel_params = [0.0, 0.0, 0.0]

    prob_a = params.probA
    prob_b = params.probB
    support_vectors_per_class = params.numberOfSupportVectorsPerClass
    n_supports, svc_support_vectors = extract_support_vectors_as_dense_tensor(
        operator.raw_operator.supportVectorClassifier
    )
    chain_coef = list(
        itertools.chain.from_iterable([coef.alpha for coef in params.coefficients])
    )
    svc_coefficients = chain_coef
    svc_rho = [-x for x in params.rho]

    op_type = "SVMClassifier"
    op_name = scope.get_unique_operator_name(op_type)
    attrs = {"name": op_name}
    attrs["kernel_type"] = svc_kernel
    attrs["kernel_params"] = svc_kernel_params
    if prob_a:
        attrs["prob_a"] = prob_a
    if prob_b:
        attrs["prob_b"] = prob_b
    attrs["vectors_per_class"] = support_vectors_per_class
    attrs["support_vectors"] = svc_support_vectors
    attrs["coefficients"] = svc_coefficients
    attrs["rho"] = svc_rho
    zipmap_attrs = {"name": scope.get_unique_operator_name("ZipMap")}
    svc_classes = params.WhichOneof("ClassLabels")
    if svc_classes == "int64ClassLabels":
        class_labels = list(int(i) for i in params.int64ClassLabels.vector)
        attrs["classlabels_ints"] = class_labels
        zipmap_attrs["classlabels_int64s"] = class_labels
    elif svc_classes == "stringClassLabels":
        class_labels = list(
            str(s).encode("utf-8") for s in params.stringClassLabels.vector
        )
        attrs["classlabels_strings"] = class_labels
        zipmap_attrs["classlabels_strings"] = class_labels
    else:
        raise ValueError("Unknown class label type")

    # For classifiers, due to the different representation
    # of classes' probabilities, we need to add some
    # operators for type conversion. It turns out that we
    # have the following topology.
    # input features ---> SupportVectorClassifier ---> label (must present)
    #                               |
    #                               '--> probability tensor

    raw_model = operator.raw_operator
    # Find label name and probability name
    proba_output_name = None
    for variable in operator.outputs:
        if raw_model.description.predictedFeatureName == variable.raw_name:
            label_output_name = variable.full_name
        if (
            raw_model.description.predictedProbabilitiesName != ""
            and raw_model.description.predictedProbabilitiesName == variable.raw_name
        ):
            proba_output_name = variable.full_name

    inputs = [variable.full_name for variable in operator.inputs]
    proba_tensor_name = scope.get_unique_variable_name("ProbabilityTensor")

    if proba_output_name is not None:
        # Add support vector classifier in terms of ONNX node with probability output
        container.add_node(
            op_type,
            inputs,
            [label_output_name, proba_tensor_name],
            op_domain="ai.onnx.ml",
            **attrs
        )

        # Add ZipMap to convert probability tensor into probability map
        container.add_node(
            "ZipMap",
            [proba_tensor_name],
            [proba_output_name],
            op_domain="ai.onnx.ml",
            **zipmap_attrs
        )
    else:
        # Add support vector classifier in terms of ONNX node
        container.add_node(
            op_type,
            inputs,
            [label_output_name, proba_tensor_name],
            op_domain="ai.onnx.ml",
            **attrs
        )


register_converter("supportVectorClassifier", convert_svm_classifier)



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import packaging.version as pv
import numpy
import pandas
import onnx
from onnx.defs import onnx_opset_version
from pyspark.ml.feature import VectorIndexer
from pyspark.ml.linalg import Vectors
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import (
    save_data_models,
    run_onnx_model,
    compare_results,
)
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlVectorIndexer(SparkMlTestCase):
    @unittest.skipIf(
        True,
        reason=(
            "discrepency, unfound values are replaced by -1 by ONNX and 0 " "by spark."
        ),
    )
    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    @unittest.skipIf(
        pv.Version(onnx.__version__) <= pv.Version("1.3"), "Need Greater Opset 9"
    )
    def test_model_vector_indexer_multi(self):
        vi = VectorIndexer(maxCategories=2, inputCol="a", outputCol="indexed")
        data = self.spark.createDataFrame(
            [
                (Vectors.dense([-1.0, 1.0, 3.1]),),
                (Vectors.dense([0.0, 5.0, 6.2]),),
                (Vectors.dense([0.0, 9.0, 3.1]),),
            ],
            ["a"],
        )
        model = vi.fit(data)
        model_onnx = convert_sparkml(
            model,
            "Sparkml VectorIndexer Multi",
            [("a", FloatTensorType([None, model.numFeatures]))],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)
        # run the model
        predicted = model.transform(data)
        expected = (
            predicted.toPandas()
            .indexed.apply(lambda x: pandas.Series(x.toArray()))
            .values
        )
        data_np = (
            data.toPandas()
            .a.apply(lambda x: pandas.Series(x.toArray()))
            .values.astype(numpy.float32)
        )
        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlVectorIndexerMulti"
        )
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(["indexed"], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)

    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    @unittest.skipIf(
        pv.Version(onnx.__version__) <= pv.Version("1.3"), "Need Greater Opset 9"
    )
    def test_model_vector_indexer_single(self):
        vi = VectorIndexer(maxCategories=3, inputCol="a", outputCol="indexed")
        data = self.spark.createDataFrame(
            [
                (Vectors.dense([-1.0]),),
                (Vectors.dense([0.0]),),
                (Vectors.dense([0.0]),),
            ],
            ["a"],
        )
        model = vi.fit(data)
        model_onnx = convert_sparkml(
            model,
            "Sparkml VectorIndexer Single",
            [("a", FloatTensorType([None, model.numFeatures]))],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)
        # run the model
        predicted = model.transform(data)
        expected = (
            predicted.toPandas()
            .indexed.apply(lambda x: pandas.Series(x.toArray()))
            .values
        )
        data_np = (
            data.toPandas()
            .a.apply(lambda x: pandas.Series(x.toArray()))
            .values.astype(numpy.float32)
        )
        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlVectorIndexerSingle"
        )
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(["indexed"], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import copy
from ....common.utils import check_input_and_output_numbers
from ....common._registration import register_shape_calculator


def calculate_concat_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N_1, C, H, W], ..., [N_n, C, H, W] ---> [N_1 + ... + N_n, C, H, W]
        2. [N, C_1, H, W], ..., [N, C_n, H, W] ---> [N, C_1 + ... + C_n, H, W]
    """
    check_input_and_output_numbers(
        operator, input_count_range=[1, None], output_count_range=[1, 1]
    )

    output_shape = copy.deepcopy(operator.inputs[0].type.shape)
    dims = []
    for variable in operator.inputs:
        if (
            variable.type.shape[0] != "None"
            and variable.type.shape[0] != output_shape[0]
        ):
            raise RuntimeError("Only dimensions along C-axis can be different")
        if (
            variable.type.shape[2] != "None"
            and variable.type.shape[2] != output_shape[2]
        ):
            raise RuntimeError("Only dimensions along C-axis can be different")
        if (
            variable.type.shape[3] != "None"
            and variable.type.shape[3] != output_shape[3]
        ):
            raise RuntimeError("Only dimensions along C-axis can be different")
        dims.append(variable.type.shape[1])

    output_shape[1] = "None" if "None" in dims else sum(dims)
    operator.outputs[0].type.shape = output_shape


register_shape_calculator("concat", calculate_concat_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ...common.data_types import Int64Type, Int64TensorType
from ...common._registration import register_converter


def convert_feature_vectorizer(scope, operator, container):
    op_type = "FeatureVectorizer"
    attrs = {"name": operator.full_name}

    inputs = []
    input_dims = []
    for variable in operator.inputs:
        if type(variable.type) in [Int64TensorType, Int64Type]:
            # We use scaler to convert integers into floats
            # because output is a single tensor and all tensor elements
            # should be in the same type.
            scaler_name = scope.get_unique_operator_name("Scaler")
            scaled_name = scope.get_unique_variable_name(variable.full_name + "_scaled")
            scaler_attrs = {"name": scaler_name, "scale": [1.0], "offset": [0.0]}
            container.add_node(
                "Scaler",
                [variable.full_name],
                [scaled_name],
                op_domain="ai.onnx.ml",
                **scaler_attrs
            )
            inputs.append(scaled_name)
        else:
            inputs.append(variable.full_name)
        # We assume feature vectorizer always combines inputs with shapes [1, C] or [C]
        input_dims.append(variable.type.shape[1])
    attrs["inputdimensions"] = input_dims

    container.add_node(
        op_type,
        inputs,
        [operator.outputs[0].full_name],
        op_domain="ai.onnx.ml",
        **attrs
    )


register_converter("featureVectorizer", convert_feature_vectorizer)



# SPDX-License-Identifier: Apache-2.0

from .....proto import onnx_proto
from ....common._registration import register_converter
from ....common._apply_operation import apply_add


def convert_add(scope, operator, container):
    if len(operator.input_full_names) == 1:
        scaler_name = scope.get_unique_variable_name(operator.full_name + "_B")
        container.add_initializer(
            scaler_name,
            onnx_proto.TensorProto.FLOAT,
            [],
            [operator.raw_operator.add.alpha],
        )
        inputs = [operator.inputs[0].full_name, scaler_name]
        broadcast = 1
        apply_add(
            scope,
            inputs,
            operator.output_full_names,
            container,
            operator_name=operator.full_name,
            broadcast=broadcast,
        )
    else:
        inputs = operator.input_full_names

        if len(inputs) == 2:
            left_tensor = operator.inputs[0].full_name
            right_tensor = operator.inputs[1].full_name
            if operator.inputs[0].type.shape != operator.inputs[1].type.shape:
                broadcast = 1
            else:
                broadcast = 0

            apply_add(
                scope,
                [left_tensor, right_tensor],
                operator.outputs[0].full_name,
                container,
                operator_name=operator.full_name,
                broadcast=broadcast,
            )
        else:
            left_tensor = operator.inputs[0].full_name
            right_tensor = operator.inputs[1].full_name

            # Sum up the first two inputs
            intermediate_tensor_name = scope.get_unique_variable_name("buffer_tensor")
            apply_add(
                scope,
                [left_tensor, right_tensor],
                intermediate_tensor_name,
                container,
                operator_name=operator.full_name,
                broadcast=1,
            )

            # Accumulate other inputs onto intermediate tensors.
            # Note that we may use the original operator's output as
            # the last intermediate tensor.
            for i in range(2, len(inputs)):
                left_tensor = intermediate_tensor_name
                right_tensor = inputs[i].full_name
                if i != len(inputs) - 1:
                    intermediate_tensor_name = scope.get_unique_variable_name(
                        "buffer_tensor"
                    )
                else:
                    intermediate_tensor_name = operator.outputs[0].full_name
                apply_add(
                    scope,
                    [left_tensor, right_tensor],
                    intermediate_tensor_name,
                    container,
                    broadcast=1,
                )


register_converter("add", convert_add)



# SPDX-License-Identifier: Apache-2.0

import numpy

from ....proto import onnx_proto
from ...common.data_types import FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._registration import register_converter, register_shape_calculator


def convert_sparkml_naive_bayes(scope, operator, container):
    op = operator.raw_operator
    model_type = op.getOrDefault("modelType")
    theta_np = numpy.array(op.theta.toArray())
    transposed_theta = numpy.transpose(theta_np)
    raw_predictions_tensor = scope.get_unique_variable_name("raw_predictions_tensor")
    if model_type == "bernoulli":
        negTheta = numpy.log(1.0 - numpy.exp(theta_np))
        thetaMinusLogTheta = numpy.transpose(theta_np - negTheta)
        ones = numpy.ones((theta_np.shape[1], 1))
        negThetaSum = numpy.matmul(negTheta, ones).flatten()
        thetaMinusLogTheta_tensor = scope.get_unique_variable_name(
            "thetaMinusLogTheta_tensor"
        )
        container.add_initializer(
            thetaMinusLogTheta_tensor,
            onnx_proto.TensorProto.FLOAT,
            list(thetaMinusLogTheta.shape),
            thetaMinusLogTheta.flatten().tolist(),
        )
        negThetaSum_tensor = scope.get_unique_variable_name("negThetaSum_tensor")
        container.add_initializer(
            negThetaSum_tensor,
            onnx_proto.TensorProto.FLOAT,
            list(negThetaSum.shape),
            negThetaSum.flatten().tolist(),
        )
        prior_tensor = scope.get_unique_variable_name("prior_tensor")
        container.add_initializer(
            prior_tensor,
            onnx_proto.TensorProto.FLOAT,
            [len(op.pi)],
            op.pi.flatten().tolist(),
        )
        probability1_output = scope.get_unique_variable_name("temp_probability")
        container.add_node(
            "MatMul",
            [operator.input_full_names[0], thetaMinusLogTheta_tensor],
            probability1_output,
            op_domain="ai.onnx",
            name=scope.get_unique_operator_name("MatMul"),
            op_version=9,
        )
        probability2_output = scope.get_unique_variable_name("temp_probability")
        container.add_node(
            "Add",
            [probability1_output, prior_tensor],
            probability2_output,
            op_domain="ai.onnx",
            name=scope.get_unique_operator_name("Add"),
            op_version=7,
        )
        container.add_node(
            "Add",
            [probability2_output, negThetaSum_tensor],
            raw_predictions_tensor,
            op_domain="ai.onnx",
            name=scope.get_unique_operator_name("Add"),
            op_version=7,
        )
    else:
        probability1_output = scope.get_unique_variable_name("temp_probability")
        theta_tensor = scope.get_unique_variable_name("theta_tensor")
        container.add_initializer(
            theta_tensor,
            onnx_proto.TensorProto.FLOAT,
            list(transposed_theta.shape),
            transposed_theta.flatten().tolist(),
        )
        container.add_node(
            "MatMul",
            [operator.input_full_names[0], theta_tensor],
            probability1_output,
            op_domain="ai.onnx",
            name=scope.get_unique_operator_name("MatMul"),
            op_version=1,
        )
        prior_tensor = scope.get_unique_variable_name("raw_predictions_tensor")
        container.add_initializer(
            prior_tensor, onnx_proto.TensorProto.FLOAT, [len(op.pi)], op.pi
        )
        container.add_node(
            "Add",
            [probability1_output, prior_tensor],
            raw_predictions_tensor,
            op_domain="ai.onnx",
            name=scope.get_unique_operator_name("Add"),
            op_version=7,
        )
    argmax_tensor = scope.get_unique_variable_name("argmax_tensor")
    container.add_node(
        "ArgMax",
        raw_predictions_tensor,
        argmax_tensor,
        op_domain="ai.onnx",
        name=scope.get_unique_operator_name("ArgMax"),
        op_version=1,
        axis=1,
    )
    container.add_node(
        "Cast",
        argmax_tensor,
        operator.output_full_names[0],
        op_domain="ai.onnx",
        name=scope.get_unique_operator_name("Cast"),
        op_version=9,
        to=1,
    )
    # Now we need to calculate Probabilities from rawPredictions
    # print('prediction:', numpy.argmax(result, 1))
    # max_log = numpy.max(result, 1).reshape(result.shape[0], 1)
    max_prediction_tensor = scope.get_unique_variable_name("max_prediction_tensor")
    container.add_node(
        "ReduceMax",
        raw_predictions_tensor,
        max_prediction_tensor,
        op_domain="ai.onnx",
        name=scope.get_unique_operator_name("ReduceMax"),
        op_version=1,
        axes=[1],
        keepdims=1,
    )
    # sub_result = result - max_log
    raw_minus_max_tensor = scope.get_unique_variable_name("raw_minus_max_tensor")
    container.add_node(
        "Sub",
        [raw_predictions_tensor, max_prediction_tensor],
        raw_minus_max_tensor,
        op_domain="ai.onnx",
        name=scope.get_unique_operator_name("Sub"),
        op_version=7,
    )
    # exp_result = numpy.exp(sub_result)
    exp_tensor = scope.get_unique_variable_name("exp_tensor")
    container.add_node(
        "Exp",
        raw_minus_max_tensor,
        exp_tensor,
        op_domain="ai.onnx",
        name=scope.get_unique_operator_name("Exp"),
        op_version=6,
    )
    # sum_log = numpy.sum(exp_result, 1).reshape(result.shape[0], 1)
    sum_prediction_tensor = scope.get_unique_variable_name("sum_prediction_tensor")
    container.add_node(
        "ReduceSum",
        exp_tensor,
        sum_prediction_tensor,
        op_domain="ai.onnx",
        name=scope.get_unique_operator_name("ReduceSum"),
        op_version=1,
        axes=[1],
        keepdims=1,
    )
    # probabilities = exp_result / sum_log
    container.add_node(
        "Div",
        [exp_tensor, sum_prediction_tensor],
        operator.output_full_names[1],
        op_domain="ai.onnx",
        name=scope.get_unique_operator_name("Div"),
        op_version=7,
    )


register_converter(
    "pyspark.ml.classification.NaiveBayesModel", convert_sparkml_naive_bayes
)


def calculate_sparkml_naive_bayes_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=2)
    check_input_and_output_types(
        operator,
        good_input_types=[FloatTensorType],
        good_output_types=[FloatTensorType, FloatTensorType],
    )
    N = operator.inputs[0].type.shape[0]
    C = operator.raw_operator.numClasses
    operator.outputs[0].type = FloatTensorType([N, 1])
    operator.outputs[1].type = FloatTensorType([N, C])


register_shape_calculator(
    "pyspark.ml.classification.NaiveBayesModel",
    calculate_sparkml_naive_bayes_output_shapes,
)



# SPDX-License-Identifier: Apache-2.0

from ....common._registration import register_converter


def convert_softmax(scope, operator, container):
    op_type = "Softmax"
    inputs = [variable.full_name for variable in operator.inputs]
    outputs = [variable.full_name for variable in operator.outputs]
    attrs = {"name": operator.full_name}
    container.add_node(op_type, inputs, outputs, **attrs)


register_converter("softmax", convert_softmax)



# SPDX-License-Identifier: Apache-2.0

from pyspark.ml.classification import LogisticRegressionModel, LinearSVCModel

from ...common._registration import register_converter, register_shape_calculator
from ...common.data_types import Int64TensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types


def convert_sparkml_linear_classifier(scope, operator, container):
    op = operator.raw_operator
    op_type = "LinearClassifier"
    attrs = {"name": scope.get_unique_operator_name(op_type)}

    if op.numClasses == 2:
        coefficients = op.coefficients.toArray().tolist()
        intercepts = [op.intercept]
        coefficients = list(map(lambda x: -1 * x, coefficients)) + coefficients
        intercepts = list(map(lambda x: -1 * x, intercepts)) + intercepts
    else:
        coefficients = op.coefficientMatrix.toArray().ravel().tolist()
        intercepts = op.interceptVector.toArray().ravel().tolist()

    if isinstance(op, LogisticRegressionModel):
        if op.numClasses > 2:
            coefficients = op.coefficientMatrix.toArray().ravel().tolist()
            intercepts = op.interceptVector.toArray().ravel().tolist()
        attrs["post_transform"] = "LOGISTIC"
    elif isinstance(op, LinearSVCModel):
        attrs["post_transform"] = "NONE"
    else:
        if op.numClasses >= 2:
            attrs["post_transform"] = "SOFTMAX"
        else:
            attrs["post_transform"] = "LOGISTIC"

    attrs["coefficients"] = coefficients
    attrs["intercepts"] = intercepts
    attrs["multi_class"] = 1 if op.numClasses >= 2 else 0
    attrs["classlabels_ints"] = list(range(0, op.numClasses))
    import pprint

    pprint.pprint(attrs)

    label_name = operator.outputs[0].full_name
    if not isinstance(operator.raw_operator, LinearSVCModel):
        probability_tensor_name = scope.get_unique_variable_name("probability_tensor")
        container.add_node(
            op_type,
            operator.inputs[0].full_name,
            [label_name, probability_tensor_name],
            op_domain="ai.onnx.ml",
            **attrs
        )

        # Make sure the probability sum is 1 over all classes
        normalizer_type = "Normalizer"
        normalizer_attrs = {
            "name": scope.get_unique_operator_name(normalizer_type),
            "norm": "L1",
        }
        container.add_node(
            normalizer_type,
            probability_tensor_name,
            operator.outputs[1].full_name,
            op_domain="ai.onnx.ml",
            **normalizer_attrs
        )
    else:
        # add a dummy output variable since onnx LinearClassifier has 2
        unused_probabilities_output = scope.get_unique_variable_name("probabilities")
        container.add_node(
            op_type,
            operator.inputs[0].full_name,
            [label_name, unused_probabilities_output],
            op_domain="ai.onnx.ml",
            **attrs
        )


register_converter(
    "pyspark.ml.classification.LogisticRegressionModel",
    convert_sparkml_linear_classifier,
)
register_converter(
    "pyspark.ml.classification.LinearSVCModel", convert_sparkml_linear_classifier
)


def calculate_linear_classifier_output_shapes(operator):
    check_input_and_output_numbers(
        operator, input_count_range=1, output_count_range=[1, 2]
    )
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType]
    )
    if len(operator.inputs[0].type.shape) != 2:
        raise RuntimeError("Input must be a [N, C]-tensor")

    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = Int64TensorType(shape=[N])
    if not isinstance(operator.raw_operator, LinearSVCModel):
        class_count = operator.raw_operator.numClasses
        operator.outputs[1].type = FloatTensorType([N, class_count])


register_shape_calculator(
    "pyspark.ml.classification.LogisticRegressionModel",
    calculate_linear_classifier_output_shapes,
)
register_shape_calculator(
    "pyspark.ml.classification.LinearSVCModel",
    calculate_linear_classifier_output_shapes,
)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ...common._registration import register_shape_calculator
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common.data_types import (
    FloatTensorType,
    Int64TensorType,
    StringTensorType,
)
from ..common import get_xgb_params, get_n_estimators_classifier


def calculate_xgboost_classifier_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=2)
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType]
    )
    N = operator.inputs[0].type.shape[0]

    xgb_node = operator.raw_operator
    params = get_xgb_params(xgb_node)
    booster = xgb_node.get_booster()
    booster.attributes()
    js_trees = booster.get_dump(with_stats=True, dump_format="json")
    ntrees = len(js_trees)
    objective = params["objective"]
    n_estimators = get_n_estimators_classifier(xgb_node, params, js_trees)
    num_class = params.get("num_class", None)

    if num_class is not None:
        ncl = num_class
        n_estimators = ntrees // ncl
    elif objective == "binary:logistic":
        ncl = 2
    else:
        ncl = ntrees // n_estimators
        if objective == "reg:logistic" and ncl == 1:
            ncl = 2
    classes = xgb_node.classes_
    if (
        np.issubdtype(classes.dtype, np.floating)
        or np.issubdtype(classes.dtype, np.integer)
        or np.issubdtype(classes.dtype, np.bool_)
    ):
        operator.outputs[0].type = Int64TensorType(shape=[N])
    else:
        operator.outputs[0].type = StringTensorType(shape=[N])
    operator.outputs[1].type = operator.outputs[1].type = FloatTensorType([N, ncl])


register_shape_calculator("XGBClassifier", calculate_xgboost_classifier_output_shapes)
register_shape_calculator("XGBRFClassifier", calculate_xgboost_classifier_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ....common._registration import register_converter


def convert_l2_normalization(scope, operator, container):
    # The first dimension is batch size, so the
    # normalization is done along the 2nd axis (indexed by 1).
    attrs = {
        "name": operator.full_name,
        "axis": 1,
        "p": 2,
    }  # Caffe normalization happens per image in one batch
    container.add_node(
        "LpNormalization",
        operator.input_full_names,
        operator.output_full_names,
        **attrs
    )


register_converter("l2normalize", convert_l2_normalization)



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import numpy
import pandas
from pyspark.ml.feature import ElementwiseProduct
from pyspark.ml.linalg import Vectors
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import (
    save_data_models,
    run_onnx_model,
    compare_results,
)
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlElementwiseProduct(SparkMlTestCase):
    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_element_wise_product(self):
        data = self.spark.createDataFrame(
            [(Vectors.dense([2.0, 1.0, 3.0]),)], ["features"]
        )
        model = ElementwiseProduct(
            scalingVec=Vectors.dense([1.0, 2.0, 3.0]),
            inputCol="features",
            outputCol="eprod",
        )
        feature_count = data.first()[0].size
        model_onnx = convert_sparkml(
            model,
            "Sparkml ElementwiseProduct",
            [("features", FloatTensorType([None, feature_count]))],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)

        # run the model
        predicted = model.transform(data)
        expected = [
            predicted.toPandas()
            .eprod.apply(lambda x: pandas.Series(x.toArray()))
            .values.astype(numpy.float32)
        ]
        data_np = (
            data.toPandas()
            .features.apply(lambda x: pandas.Series(x.toArray()))
            .values.astype(numpy.float32)
        )
        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlElementwiseProduct"
        )
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(["eprod"], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from .....proto import onnx_proto
from ....common._apply_operation import apply_concat, apply_split
from ....common._registration import register_converter
from .SimpleRNN import extract_rnn_activation_info
from .Reshape import apply_reshape


def convert_bidirectional_lstm(scope, operator, container):
    # The LSTM inputs are feature vector (X), initial forward hidden state (h_init), initial backward hidden
    # state (h_init_rev), initial forward cell state (c_init), and initial backward cell stat (c_init_rev)e.
    # Because of the shape differences between ONNX LSTM and CoreML LSTM, the bidirectional LSTM conversion
    # is not straightforward. See some visualizations below for details.
    #
    # Symbols:
    #
    #  C_in: input feature length
    #  C_out: output feature length
    #  S: sequence length. For example, it can be the number of tokens in a sentence.
    #
    #  X: input features of CoreML LSTM. Its shape is [S, C_in].
    #  h_init: initial forward hidden state in CoreML. Its shape is [1, C_out].
    #  c_init: initial forward cell state in CoreML. Its shape is [1, C_out].
    #  h_init_rev: initial backward hidden state in CoreML. Its shape is [1, C_out].
    #  c_init_rev: initial backward cell state in CoreML. Its shape is [1, C_out].
    #  Y: CoreML LSTM's output. It can be [S, C_out] (if sequence_output is on) or [1, C_out] (if sequence_output is off)
    #  Y_h: CoreML LSTM's last foward hidden state. Its shape is [1, C_out].
    #  Y_c: CoreML LSTM's last forward cell state. Its shape is [1, C_out].
    #  Y_h_rev: CoreML LSTM's last backward hidden state. Its shape is [1, C_out].
    #  Y_c_rev: CoreML LSTM's last backward cell state. Its shape is [1, C_out].
    #
    #  X': input features of ONNX LSTM
    #  h_init': initial (forward and backward) hidden states of ONNX LSTM
    #  c_init': initial (forward and backward) hidden states of ONNX LSTM
    #  Y': ONNX LSTM's output
    #  Y_h': ONNX LSTM's last (forward and backward) hidden states
    #  Y_c': ONNX LSTM's last (forward and backward) cell states
    #
    # Computational graph of CoreML bi-directional LSTM (if sequence_output is on):
    #
    #      X [S, C_in]  h_init [1, C_out]  c_init [1, C_out]  h_init_rev [1, C_out]  c_init_rev [1, C_out]
    #      |               |                    |                |                      |
    #      |               '-----------------.  |                |                      |
    #      |                                 |  |                |   .------------------'
    #      '--------------------------------.|  |                |   |
    #                                       ||  |                |   |
    #                                       vv  v                v   v
    #                                       CoreML Bi-directional LSTM
    #                                       ||  |                |   |
    #      .--------------------------------'|  |                |   '------------------.
    #      |                                 |  |                |                      |
    #      |              .------------------'  |                |                      |
    #      |              |                     |                |                      |
    #      v              v                     v                v                      v
    #      Y [S, 2*C_out] Y_h [1, C_out]     Y_c [1, C_out]    Y_h_rev [1, C_out]     Y_c_rev [1, C_out]
    #
    # Computational graph of CoreML bi-directional LSTM in ONNX (if sequence_output is on):
    #
    #      X [S, C_in]  h_init [1, C_out]  h_init_rev [1, C_out] c_init [1, C_out]   c_init_rev [1, C_out]
    #      |               |                    |                  |                      |
    #      |               |                    |                  |                      |
    #      |               '-------.    .-------'                  '--------.    .--------'
    #      |                       |    |                                   |    |
    #      v                       v    v                                   v    v
    #   Reshape                    Concate                                  Concate
    #      |                          |                                        |
    #      v                          v                                        v
    #      X' [S, 1, C_in]        _h_init_ [2, C_out]                      _c_init_ [2, C_out]
    #      |                          |                                        |
    #      |                          v                                        v
    #      |                       Reshape                                  Reshape
    #      |                          |                                        |
    #      |                          v                                        V
    #      |                       h_init' [2, 1, C_out]                    c_init' [2, 1, Cout]
    #      |                          |                                        |
    #      |                          '-----------------.       .--------------'
    #      '----------------------------------.         |       |
    #                                         |         |       |
    #                                         v         v       v
    #                                       ONNX Bi-directional LSTM
    #                                         |         |       |
    #      .----------------------------------'         |       '--------------------.
    #      |                                            |                            |
    #      v                                            v                            v
    #      Y' [S, 2, 1, C_out]                           Y_h' [2, 1, C_out]           Y_c' [2, 1, C_out]
    #      |                                            |                            |
    #      v                                            v                            v
    #   Reshape                                      Reshape                      Reshape
    #      |                                            |                            |
    #      v                                            v                            v
    #      Y  [S, 2*C_out]                             _Y_h_' [2, C_out]            _Y_c_' [2, C_out]
    #                                                   |                            |
    #                                                   v                            v
    #                                                 Split                        Split
    #                                                 |   |                        |   |
    #                     .---------------------------'   |      .-----------------'   |
    #                     |                     .---------'      |                     |
    #                     |                     |                |                     |
    #                     v                     v                v                     v
    #                    Y_h [1, C_out]     Y_h_rev [1, C_out]  Y_c [1, C_out]       Y_c_rev [1, C_out]
    #
    # Computational graph of CoreML bi-directional LSTM (if sequence_output is off):
    #
    #      X [S, C_in]  h_init [1, C_out]  c_init [1, C_out]  h_init_rev [1, C_out]  c_init_rev [1, C_out]
    #      |               |                    |                |                      |
    #      |               '-----------------.  |                |                      |
    #      |                                 |  |                |   .------------------'
    #      '--------------------------------.|  |                |   |
    #                                       ||  |                |   |
    #                                       vv  v                v   v
    #                                       CoreML Bi-directional LSTM
    #                                       ||  |                |   |
    #      .--------------------------------'|  |                |   '------------------.
    #      |                                 |  |                |                      |
    #      |              .------------------'  |                |                      |
    #      |              |                     |                |                      |
    #      v              v                     v                v                      v
    #      Y [1, 2*C_out] Y_h [1, C_out]     Y_c [1, C_out]    Y_h_rev [1, C_out]     Y_c_rev [1, C_out]
    #
    # Computational graph of CoreML bi-directional LSTM in ONNX (if sequence_output is off):
    #
    #      X [S, C_in]  h_init [1, C_out]  h_init_rev [1, C_out] c_init [1, C_out]   c_init_rev [1, C_out]
    #      |               |                    |                  |                      |
    #      |               |                    |                  |                      |
    #      |               '-------.    .-------'                  '--------.    .--------'
    #      |                       |    |                                   |    |
    #      v                       v    v                                   v    v
    #   Reshape                    Concate                                  Concate
    #      |                          |                                        |
    #      v                          v                                        v
    #      X' [S, 1, C_in]        _h_init_ [2, C_out]                      _c_init_ [2, C_out]
    #      |                          |                                        |
    #      |                          v                                        v
    #      |                       Reshape                                  Reshape
    #      |                          |                                        |
    #      |                          v                                        V
    #      |                       h_init' [2, 1, C_out]                    c_init' [2, 1, Cout]
    #      |                          |                                        |
    #      |                          '-----------------.       .--------------'
    #      '----------------------------------.         |       |
    #                                         |         |       |
    #                                         v         v       v
    #                                       ONNX Bi-directional LSTM
    #                                         |         |       |
    #      .----------------------------------'         |       '--------------------.
    #      |                                            |                            |
    #      v                                            v                            v
    #      Y' [S, 2, 1, C_in]          .-------------- Y_h' [2, 1, C_out]           Y_c' [2, 1, C_out]
    #   (useless output)               |                |                            |
    #                                  v                v                            v
    #                               Reshape          Reshape                      Reshape
    #                                  |                |                            |
    #                                  |                v                            v
    #   .------------------------------'              _Y_h_' [2, C_out]            _Y_c_' [2, C_out]
    #   |                                               |                            |
    #   |                                               v                            v
    #   |                                             Split                        Split
    #   |                                             |   |                        |   |
    #   |                 .---------------------------'   |      .-----------------'   |
    #   |                 |                     .---------'      |                     |
    #   |                 |                     |                |                     |
    #   v                 v                     v                v                     v
    #   Y  [1, 2*C_out]   Y_h [1, C_out]     Y_h_rev [1, C_out]  Y_c [1, C_out]       Y_c_rev [1, C_out]

    params = operator.raw_operator.biDirectionalLSTM
    lstm_params = params.params
    lstm_weights = params.weightParams
    input_size = params.inputVectorSize
    hidden_size = params.outputVectorSize

    lstm_op_name = scope.get_unique_operator_name("LSTM")
    lstm_attrs = {"name": lstm_op_name}
    lstm_inputs = []
    lstm_outputs = []

    lstm_x_reshape_name = scope.get_unique_variable_name(lstm_op_name + "_X_reshape")

    apply_reshape(
        scope,
        operator.inputs[0].full_name,
        lstm_x_reshape_name,
        container,
        desired_shape=[-1, 1, input_size],
    )
    lstm_inputs.append(lstm_x_reshape_name)

    # Handle LSTM's weight matrices
    matrices_w_forward = np.concatenate(
        [
            lstm_weights[0].inputGateWeightMatrix.floatValue,
            lstm_weights[0].outputGateWeightMatrix.floatValue,
            lstm_weights[0].forgetGateWeightMatrix.floatValue,
            lstm_weights[0].blockInputWeightMatrix.floatValue,
        ]
    )
    matrices_w_backward = np.concatenate(
        [
            lstm_weights[1].inputGateWeightMatrix.floatValue,
            lstm_weights[1].outputGateWeightMatrix.floatValue,
            lstm_weights[1].forgetGateWeightMatrix.floatValue,
            lstm_weights[1].blockInputWeightMatrix.floatValue,
        ]
    )
    matrices_w_name = scope.get_unique_variable_name(lstm_op_name + "_W")
    container.add_initializer(
        matrices_w_name,
        onnx_proto.TensorProto.FLOAT,
        [2, 4 * hidden_size, input_size],
        np.concatenate([matrices_w_forward, matrices_w_backward]),
    )
    lstm_inputs.append(matrices_w_name)

    # Handle LSTM's recursion matrices
    matrices_r_forward = np.concatenate(
        [
            lstm_weights[0].inputGateRecursionMatrix.floatValue,
            lstm_weights[0].outputGateRecursionMatrix.floatValue,
            lstm_weights[0].forgetGateRecursionMatrix.floatValue,
            lstm_weights[0].blockInputRecursionMatrix.floatValue,
        ]
    )
    matrices_r_backward = np.concatenate(
        [
            lstm_weights[1].inputGateRecursionMatrix.floatValue,
            lstm_weights[1].outputGateRecursionMatrix.floatValue,
            lstm_weights[1].forgetGateRecursionMatrix.floatValue,
            lstm_weights[1].blockInputRecursionMatrix.floatValue,
        ]
    )
    matrices_r_name = scope.get_unique_variable_name(lstm_op_name + "_R")
    container.add_initializer(
        matrices_r_name,
        onnx_proto.TensorProto.FLOAT,
        [2, 4 * hidden_size, hidden_size],
        np.concatenate([matrices_r_forward, matrices_r_backward]),
    )
    lstm_inputs.append(matrices_r_name)

    # Handle bias vectors
    vectors_b = np.zeros(shape=(2, 8, hidden_size))
    if lstm_params.hasBiasVectors:
        vectors_b[0, 0, :] = lstm_weights[0].inputGateBiasVector.floatValue
        vectors_b[0, 1, :] = lstm_weights[0].outputGateBiasVector.floatValue
        vectors_b[0, 2, :] = lstm_weights[0].forgetGateBiasVector.floatValue
        vectors_b[0, 3, :] = lstm_weights[0].blockInputBiasVector.floatValue
        vectors_b[1, 0, :] = lstm_weights[1].inputGateBiasVector.floatValue
        vectors_b[1, 1, :] = lstm_weights[1].outputGateBiasVector.floatValue
        vectors_b[1, 2, :] = lstm_weights[1].forgetGateBiasVector.floatValue
        vectors_b[1, 3, :] = lstm_weights[1].blockInputBiasVector.floatValue
    if lstm_params.forgetBias:
        # One may think we should do something like b[0, 2, :] += 1. and b[1, 2, :] += 1.,
        # but it's not correct as CoreML has added 1 into those bias vectors.
        pass
    if lstm_params.hasBiasVectors or lstm_params.forgetBias:
        vectors_b_name = scope.get_unique_variable_name(lstm_op_name + "_B")
        container.add_initializer(
            vectors_b_name,
            onnx_proto.TensorProto.FLOAT,
            [2, 8 * hidden_size],
            vectors_b.flatten(),
        )
        lstm_inputs.append(vectors_b_name)
    else:
        lstm_inputs.append("")

    # Due to the position sensitivity in ONNX argument parsing,
    # we add an empty string for the non-existing
    # sequence length
    lstm_inputs.append("")

    # Handle initial hidden state if necessary
    if len(operator.inputs) > 1:
        lstm_h_init_name = scope.get_unique_variable_name(lstm_op_name + "_h_init")
        apply_concat(
            scope,
            [operator.inputs[1].full_name, operator.inputs[3].full_name],
            lstm_h_init_name,
            container,
            axis=0,
        )

        lstm_h_init_reshape_name = scope.get_unique_variable_name(
            lstm_op_name + "_h_init_reshape"
        )
        apply_reshape(
            scope,
            lstm_h_init_name,
            lstm_h_init_reshape_name,
            container,
            desired_shape=[2, 1, hidden_size],
        )

        # Add zero initializers to forward and backward initial
        # hidden states so that they become optional
        container.add_initializer(
            operator.inputs[1].full_name,
            onnx_proto.TensorProto.FLOAT,
            operator.inputs[1].type.shape,
            np.zeros(shape=operator.inputs[1].type.shape).flatten(),
        )
        container.add_initializer(
            operator.inputs[3].full_name,
            onnx_proto.TensorProto.FLOAT,
            operator.inputs[3].type.shape,
            np.zeros(shape=operator.inputs[3].type.shape).flatten(),
        )
        lstm_inputs.append(lstm_h_init_reshape_name)
    else:
        lstm_inputs.append("")

    # Handle initial cell state if needed
    if len(operator.inputs) > 2:
        lstm_c_init_name = scope.get_unique_variable_name(lstm_op_name + "_c_init")
        apply_concat(
            scope,
            [operator.inputs[2].full_name, operator.inputs[4].full_name],
            lstm_c_init_name,
            container,
            axis=0,
        )

        # Reshape the Cell state so that ONNX LSTM can accept it
        lstm_c_init_reshape_name = scope.get_unique_variable_name(
            lstm_op_name + "_c_init_reshape"
        )
        apply_reshape(
            scope,
            lstm_c_init_name,
            lstm_c_init_reshape_name,
            container,
            desired_shape=[2, 1, hidden_size],
        )
        lstm_inputs.append(lstm_c_init_reshape_name)

        # Add zero initializers to forward and backward initial
        # cell states so that they become optional
        container.add_initializer(
            operator.inputs[2].full_name,
            onnx_proto.TensorProto.FLOAT,
            operator.inputs[2].type.shape,
            np.zeros(shape=operator.inputs[2].type.shape).flatten(),
        )
        container.add_initializer(
            operator.inputs[4].full_name,
            onnx_proto.TensorProto.FLOAT,
            operator.inputs[4].type.shape,
            np.zeros(shape=operator.inputs[4].type.shape).flatten(),
        )
    else:
        lstm_inputs.append("")

    # Handle peephole vectors if necessary
    if lstm_params.hasPeepholeVectors:
        p_forward = np.concatenate(
            [
                lstm_weights[0].inputGatePeepholeVector.floatValue,
                lstm_weights[0].outputGatePeepholeVector.floatValue,
                lstm_weights[0].forgetGatePeepholeVector.floatValue,
            ]
        )
        p_backward = np.concatenate(
            [
                lstm_weights[1].inputGatePeepholeVector.floatValue,
                lstm_weights[1].outputGatePeepholeVector.floatValue,
                lstm_weights[1].forgetGatePeepholeVector.floatValue,
            ]
        )
        p_name = scope.get_unique_variable_name(lstm_op_name + "_P")
        container.add_initializer(
            p_name,
            onnx_proto.TensorProto.FLOAT,
            [2, 3 * hidden_size],
            np.concatenate([p_forward, p_backward]),
        )
        lstm_inputs.append(p_name)
    else:
        lstm_inputs.append("")

    # Parse activation functions and add them into ONNX LSTM's attribute dictionary
    activation_types = []
    alphas = []
    betas = []
    for activation in params.activationsForwardLSTM:
        activation_type, alpha, beta = extract_rnn_activation_info(activation)
        activation_types.append(activation_type.encode("utf-8"))
        if alpha is not None:
            alphas.append(alpha)
        if beta is not None:
            betas.append(beta)
    for activation in params.activationsBackwardLSTM:
        activation_type, alpha, beta = extract_rnn_activation_info(activation)
        activation_types.append(activation_type.encode("utf-8"))
        if alpha is not None:
            alphas.append(alpha)
        if beta is not None:
            betas.append(beta)
    lstm_attrs["activations"] = activation_types
    if alphas:
        lstm_attrs["activation_alpha"] = alphas
    if betas:
        lstm_attrs["activation_beta"] = betas

    # Add more attributes
    lstm_attrs["direction"] = "bidirectional"
    lstm_attrs["hidden_size"] = hidden_size
    lstm_attrs["clip"] = float(lstm_params.cellClipThreshold)
    lstm_attrs["input_forget"] = lstm_params.coupledInputAndForgetGate

    # Set up version-dependent attributes
    if container.target_opset < 7:
        lstm_attrs["output_sequence"] = lstm_params.sequenceOutput
        op_version = 1
    else:
        op_version = 7

    # Create the major ONNX LSTM operator. We assign a tensor name
    # to each output of LSTM. However, variables can be
    # undefined in some cases. For example, when output_sequence=False,
    # the first output is not meaningful.
    lstm_y_name = scope.get_unique_variable_name(lstm_op_name + "_Y")
    lstm_y_h_name = scope.get_unique_variable_name(lstm_op_name + "_Y_h")
    lstm_y_c_name = scope.get_unique_variable_name(lstm_op_name + "_Y_c")
    lstm_outputs.extend([lstm_y_name, lstm_y_h_name, lstm_y_c_name])
    container.add_node(
        "LSTM", lstm_inputs, lstm_outputs, op_version=op_version, **lstm_attrs
    )

    # Create post-processing operators for converting ONNX LSTM outputs to CoreML ones
    if lstm_params.sequenceOutput:
        apply_reshape(
            scope,
            lstm_y_name,
            operator.outputs[0].full_name,
            container,
            desired_shape=[-1, 2 * hidden_size],
        )

        if len(operator.outputs) > 1:
            lstm_y_h_reshape_name = scope.get_unique_variable_name(
                lstm_op_name + "_Y_h_reshape"
            )
            apply_reshape(
                scope,
                lstm_y_h_name,
                lstm_y_h_reshape_name,
                container,
                desired_shape=[2, hidden_size],
            )

            apply_split(
                scope,
                lstm_y_h_reshape_name,
                [operator.outputs[1].full_name, operator.outputs[3].full_name],
                container,
                split=[1, 1],
                axis=0,
            )
    else:
        # Here we ignore ONNX RNN's first output because it's useless.
        # The second output of ONNX LSTM will be used to
        # generate the first and the second outputs of CoreML LSTM.

        # Directly reshape ONNX LSTM's 2nd output to CoreML LSTM's 1st output.
        apply_reshape(
            scope,
            lstm_y_h_name,
            operator.outputs[0].full_name,
            container,
            desired_shape=[1, 2 * hidden_size],
        )

        if len(operator.outputs) > 1:
            lstm_y_reshape_name = scope.get_unique_variable_name(
                lstm_op_name + "_Y_reshape"
            )
            apply_reshape(
                scope,
                lstm_y_h_name,
                lstm_y_reshape_name,
                container,
                desired_shape=[2, hidden_size],
            )

            apply_split(
                scope,
                lstm_y_reshape_name,
                [operator.outputs[1].full_name, operator.outputs[3].full_name],
                container,
                split=[1, 1],
                axis=0,
            )

    # Output cell state if necessary
    if len(operator.outputs) > 2:
        lstm_y_c_reshape_name = scope.get_unique_variable_name(
            lstm_op_name + "_Y_c_reshape"
        )
        apply_reshape(
            scope,
            lstm_y_c_name,
            lstm_y_c_reshape_name,
            container,
            desired_shape=[2, hidden_size],
        )

        apply_split(
            scope,
            lstm_y_c_reshape_name,
            [operator.outputs[2].full_name, operator.outputs[4].full_name],
            container,
            split=[1, 1],
            axis=0,
        )


register_converter("biDirectionalLSTM", convert_bidirectional_lstm)



# SPDX-License-Identifier: Apache-2.0

import copy
from ...common._registration import register_shape_calculator
from ...common.data_types import FloatTensorType, Int64TensorType, StringTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculate_array_feature_extractor_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C] ---> [N, C']

    C' is the number of extracted features.
    """
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType, StringTensorType]
    )

    N = operator.inputs[0].type.shape[0]
    extracted_feature_number = len(
        operator.raw_operator.arrayFeatureExtractor.extractIndex
    )

    # Save doc_string before over-writing by us
    doc_string = operator.outputs[0].type.doc_string
    operator.outputs[0].type = copy.deepcopy(operator.inputs[0].type)
    operator.outputs[0].type.shape = [N, extracted_feature_number]
    # Assign correct doc_string to the output
    operator.outputs[0].type.doc_string = doc_string


register_shape_calculator(
    "arrayFeatureExtractor", calculate_array_feature_extractor_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

import copy
import math
from onnx import onnx_pb as onnx_proto

import numpy

from ...common.data_types import FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._apply_operation import apply_matmul, apply_mul
from ...common._registration import register_converter, register_shape_calculator


def convert_sparkml_dct(scope, operator, container):
    # op = operator.raw_operator
    # inverse = op.getInverse()
    K = operator.inputs[0].type.shape[1]
    two_sqrt_n = math.sqrt(2.0 / K)
    sqrt_n = math.sqrt(1.0 / K)
    cosine_matrix = numpy.array(
        [
            [math.cos(math.pi * (n + 0.5) * k / K) for n in range(0, K)]
            for k in range(0, K)
        ]
    ).transpose()
    cosine_tensor = scope.get_unique_variable_name("cosine_tensor")
    container.add_initializer(
        cosine_tensor,
        onnx_proto.TensorProto.FLOAT,
        cosine_matrix.shape,
        list(cosine_matrix.flatten()),
    )
    matmul_result = scope.get_unique_variable_name("matmul_tensor")
    apply_matmul(
        scope, [operator.inputs[0].full_name, cosine_tensor], matmul_result, container
    )
    scale_vector = [sqrt_n if (k == 0) else two_sqrt_n for k in range(0, K)]
    scale_tensor = scope.get_unique_variable_name("scale_tensor")
    container.add_initializer(
        scale_tensor, onnx_proto.TensorProto.FLOAT, [1, len(scale_vector)], scale_vector
    )
    apply_mul(
        scope,
        [matmul_result, scale_tensor],
        operator.output_full_names,
        container,
        axis=1,
        broadcast=True,
    )


register_converter("pyspark.ml.feature.DCT", convert_sparkml_dct)


def calculate_sparkml_dct_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])
    operator.outputs[0].type = copy.deepcopy(operator.inputs[0].type)


register_shape_calculator("pyspark.ml.feature.DCT", calculate_sparkml_dct_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter
from ...common._registration import register_shape_calculator
from ...common.utils import check_input_and_output_numbers
from ...common.data_types import FloatTensorType, Int64TensorType


def convert_sparkml_vector_assembler(scope, operator, container):
    container.add_node(
        "Concat",
        [s for s in operator.input_full_names],
        operator.outputs[0].full_name,
        name=scope.get_unique_operator_name("Concat"),
        op_version=4,
        axis=1,
    )


register_converter(
    "pyspark.ml.feature.VectorAssembler", convert_sparkml_vector_assembler
)


def calculate_vector_assembler_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)

    # Sum up the rank 1 (length of input vector) from each input.
    C = sum([input.type.shape[1] for input in operator.inputs])
    N = operator.inputs[0].type.shape[0]
    col_type = operator.inputs[0].type
    if isinstance(col_type, FloatTensorType):
        col_type = FloatTensorType([N, C])
    elif isinstance(col_type, Int64TensorType):
        col_type = Int64TensorType([N, C])
    else:
        raise TypeError("Unsupported input type")
    operator.outputs[0].type = col_type


register_shape_calculator(
    "pyspark.ml.feature.VectorAssembler", calculate_vector_assembler_shapes
)



# SPDX-License-Identifier: Apache-2.0

"""
Common function to converters and shape calculators.
"""
import json


def get_xgb_params(xgb_node):
    """
    Retrieves parameters of a model.
    """
    if hasattr(xgb_node, "kwargs"):
        # XGBoost >= 0.7
        params = xgb_node.get_xgb_params()
    else:
        # XGBoost < 0.7
        params = xgb_node.__dict__
    if hasattr("xgb_node", "save_config"):
        config = json.loads(xgb_node.save_config())
    else:
        config = json.loads(xgb_node.get_booster().save_config())
    params = {k: v for k, v in params.items() if v is not None}
    num_class = int(config["learner"]["learner_model_param"]["num_class"])
    if num_class > 0:
        params["num_class"] = num_class
    if "n_estimators" not in params and hasattr(xgb_node, "n_estimators"):
        # xgboost >= 1.0.2
        if xgb_node.n_estimators is not None:
            params["n_estimators"] = xgb_node.n_estimators
    if "base_score" in config["learner"]["learner_model_param"]:
        bs = float(config["learner"]["learner_model_param"]["base_score"])
        # xgboost >= 2.0
        params["base_score"] = bs

    bst = xgb_node.get_booster()
    if hasattr(bst, "best_ntree_limit"):
        params["best_ntree_limit"] = bst.best_ntree_limit
    if "gradient_booster" in config["learner"]:
        gbp = config["learner"]["gradient_booster"]["gbtree_model_param"]
        if "num_trees" in gbp:
            params["best_ntree_limit"] = int(gbp["num_trees"])
    return params


def get_n_estimators_classifier(xgb_node, params, js_trees):
    if "n_estimators" not in params:
        num_class = params.get("num_class", 0)
        if num_class == 0:
            return len(js_trees)
        return len(js_trees) // num_class
    return params["n_estimators"]



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter


def convert_imputer(scope, operator, container):
    op_type = "Imputer"
    attrs = {"name": operator.full_name}
    imputer = operator.raw_operator.imputer
    if imputer.HasField("replaceDoubleValue"):
        attrs["replaced_value_float"] = imputer.replaceDoubleValue
    elif imputer.HasField("replaceInt64Value"):
        attrs["replaced_value_int64"] = imputer.replaceInt64Value
    if imputer.HasField("imputedDoubleArray"):
        attrs["imputed_value_floats"] = imputer.imputedDoubleArray.vector
    elif imputer.HasField("imputedInt64Array"):
        attrs["imputed_value_int64s"] = imputer.imputedInt64Array.vector
    container.add_node(
        op_type,
        operator.input_full_names,
        operator.output_full_names,
        op_domain="ai.onnx.ml",
        **attrs
    )


register_converter("imputer", convert_imputer)



# SPDX-License-Identifier: Apache-2.0

from ....common._registration import register_converter


def convert_reorganize_data(scope, operator, container):
    from coremltools.proto.NeuralNetwork_pb2 import ReorganizeDataLayerParams as Params

    params = operator.raw_operator.reorganizeData
    if params.mode == Params.DEPTH_TO_SPACE:
        op_type = "DepthToSpace"
    elif params.mode == Params.SPACE_TO_DEPTH:
        op_type = "SpaceToDepth"
    else:
        raise ValueError("Unsupported reorganization mode {0}".format(params.mode))

    op_name = scope.get_unique_operator_name(op_type)
    attrs = {"name": op_name, "blocksize": params.blockSize}
    container.add_node(
        op_type, operator.input_full_names, operator.output_full_names, **attrs
    )


register_converter("reorganizeData", convert_reorganize_data)



# SPDX-License-Identifier: Apache-2.0

from uuid import uuid4
import xgboost
import onnx
from ..common.onnx_ex import get_maximum_opset_supported
from ..common._topology import convert_topology
from ._parse import parse_xgboost, WrappedBooster

# Invoke the registration of all our converters and shape calculators
# from . import shape_calculators


def convert(
    model,
    name=None,
    initial_types=None,
    doc_string="",
    target_opset=None,
    targeted_onnx=onnx.__version__,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
):
    """
    This function produces an equivalent ONNX model of the given xgboost model.

    :param model: A xgboost model
    :param initial_types: a python list. Each element is a tuple
        of a variable name and a type defined in data_types.py
    :param name: The name of the graph (type: GraphProto)
        in the produced ONNX model (type: ModelProto)
    :param doc_string: A string attached onto the produced ONNX model
    :param target_opset: number, for example, 7 for ONNX 1.2, and 8 for ONNX 1.3.
    :param targeted_onnx: A string (for example, '1.1.2' and '1.2')
        used to specify the targeted ONNX version of the
        produced model. If ONNXMLTools cannot find a compatible
        ONNX python package, an error may be thrown.
    :param custom_conversion_functions: a dictionary for
        specifying the user customized conversion function
    :param custom_shape_calculators: a dictionary for
        specifying the user customized shape calculator
    :return: An ONNX model (type: ModelProto) which is equivalent to the input xgboost model
    """
    if initial_types is None:
        raise ValueError(
            "Initial types are required. See usage of convert(...) in \
                           onnxmltools.convert.xgboost.convert for details"
        )
    if name is None:
        name = str(uuid4().hex)

    if isinstance(model, xgboost.Booster):
        model = WrappedBooster(model)
    target_opset = target_opset if target_opset else get_maximum_opset_supported()
    topology = parse_xgboost(
        model,
        initial_types,
        target_opset,
        custom_conversion_functions,
        custom_shape_calculators,
    )
    topology.compile()
    onnx_model = convert_topology(
        topology, name, doc_string, target_opset, targeted_onnx
    )
    opsets = {d.domain: d.version for d in onnx_model.opset_import}
    if "" in opsets and opsets[""] < 9 and target_opset >= 9:
        # a bug?
        opsets[""] = 9
        del onnx_model.opset_import[:]
        for k, v in opsets.items():
            opset = onnx_model.opset_import.add()
            opset.domain = k
            opset.version = v
    return onnx_model



# SPDX-License-Identifier: Apache-2.0

from pyspark.ml.classification import MultilayerPerceptronClassificationModel

from ...common._registration import register_converter, register_shape_calculator
from ...common.data_types import Int64TensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._topology import Operator, Scope, ModelComponentContainer
from ....proto import onnx_proto
from typing import List
import numpy as np


def convert_sparkml_mlp_classifier(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    op: MultilayerPerceptronClassificationModel = operator.raw_operator
    layers: List[int] = op.getLayers()
    weights: np.ndarray = op.weights.toArray()

    offset = 0

    input: str
    for i in range(len(layers) - 1):
        weight_matrix = weights[offset : offset + layers[i] * layers[i + 1]].reshape(
            [layers[i], layers[i + 1]]
        )
        offset += layers[i] * layers[i + 1]
        bias_vector = weights[offset : offset + layers[i + 1]]
        offset += layers[i + 1]

        if i == 0:
            input = operator.inputs[0].full_name

        weight_variable = scope.get_unique_variable_name("w")
        container.add_initializer(
            weight_variable,
            onnx_proto.TensorProto.FLOAT,
            weight_matrix.shape,
            weight_matrix.flatten().astype(np.float32),
        )

        bias_variable = scope.get_unique_variable_name("b")
        container.add_initializer(
            bias_variable,
            onnx_proto.TensorProto.FLOAT,
            bias_vector.shape,
            bias_vector.astype(np.float32),
        )

        gemm_output_variable = scope.get_unique_variable_name("gemm_output")
        container.add_node(
            op_type="Gemm",
            inputs=[input, weight_variable, bias_variable],
            outputs=[gemm_output_variable],
            op_version=7,
            name=scope.get_unique_operator_name("Gemm"),
        )

        if i == len(layers) - 2:
            container.add_node(
                op_type="Softmax",
                inputs=[gemm_output_variable],
                outputs=[operator.outputs[1].full_name],
                op_version=1,
                name=scope.get_unique_operator_name("Softmax"),
            )
        else:
            input = scope.get_unique_variable_name("activation_output")
            container.add_node(
                op_type="Sigmoid",
                inputs=[gemm_output_variable],
                outputs=[input],
                op_version=1,
                name=scope.get_unique_operator_name("Sigmoid"),
            )

    container.add_node(
        "ArgMax",
        [operator.outputs[1].full_name],
        [operator.outputs[0].full_name],
        name=scope.get_unique_operator_name("ArgMax"),
        axis=1,
        keepdims=0,
    )


register_converter(
    "pyspark.ml.classification.MultilayerPerceptronClassificationModel",
    convert_sparkml_mlp_classifier,
)


def calculate_mlp_classifier_output_shapes(operator: Operator):
    op: MultilayerPerceptronClassificationModel = operator.raw_operator

    check_input_and_output_numbers(
        operator, input_count_range=1, output_count_range=[1, 2]
    )
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType]
    )

    if len(operator.inputs[0].type.shape) != 2:
        raise RuntimeError("Input must be a [N, C]-tensor")

    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = Int64TensorType(shape=[N])
    class_count = op.numClasses
    operator.outputs[1].type = FloatTensorType([N, class_count])


register_shape_calculator(
    "pyspark.ml.classification.MultilayerPerceptronClassificationModel",
    calculate_mlp_classifier_output_shapes,
)



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import numpy
from pyspark.ml.feature import OneHotEncoder
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import (
    save_data_models,
    run_onnx_model,
    compare_results,
)
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlOneHotEncoder(SparkMlTestCase):
    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_model_onehot_encoder_1(self):
        """
        Testing ONNX conversion for Spark OneHotEncoder when handleInvalid
        is set to "error" and dropLast set to False.
        """
        encoder = OneHotEncoder(
            inputCols=["index1", "index2"],
            outputCols=["index1Vec", "index2Vec"],
            handleInvalid="error",
            dropLast=False,
        )
        data = self.spark.createDataFrame(
            [
                (
                    0.0,
                    5.0,
                ),
                (
                    1.0,
                    4.0,
                ),
                (
                    2.0,
                    3.0,
                ),
                (
                    2.0,
                    2.0,
                ),
                (
                    0.0,
                    1.0,
                ),
                (
                    2.0,
                    0.0,
                ),
            ],
            ["index1", "index2"],
        )
        model = encoder.fit(data)
        model_onnx = convert_sparkml(
            model,
            "Sparkml OneHotEncoder",
            [
                ("index1", FloatTensorType([None, 1])),
                ("index2", FloatTensorType([None, 1])),
            ],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)
        self.assertTrue(model_onnx.graph.node is not None)
        # run the model
        predicted = model.transform(data)
        data_np = {
            "index1": data.select("index1").toPandas().values.astype(numpy.float32),
            "index2": data.select("index2").toPandas().values.astype(numpy.float32),
        }

        predicted_np_1 = (
            predicted.select("index1Vec")
            .toPandas()
            .index1Vec.apply(lambda x: x.toArray())
            .values
        )
        predicted_np_2 = (
            predicted.select("index2Vec")
            .toPandas()
            .index2Vec.apply(lambda x: x.toArray())
            .values
        )
        expected = {
            "index1Vec": numpy.asarray(predicted_np_1.tolist()),
            "index2Vec": numpy.asarray(predicted_np_2.tolist()),
        }

        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlOneHotEncoder"
        )
        onnx_model_path = paths[-1]

        output_names = ["index1Vec", "index2Vec"]
        output, output_shapes = run_onnx_model(output_names, data_np, onnx_model_path)
        actual_output = dict(zip(output_names, output))

        compare_results(expected["index1Vec"], actual_output["index1Vec"], decimal=5)
        compare_results(expected["index2Vec"], actual_output["index2Vec"], decimal=5)

    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_model_onehot_encoder_2(self):
        """
        Testing ONNX conversion for Spark OneHotEncoder when handleInvalid
        is set to "keep" and dropLast set to True.
        """
        encoder = OneHotEncoder(
            inputCols=["index1", "index2"],
            outputCols=["index1Vec", "index2Vec"],
            handleInvalid="keep",
            dropLast=True,
        )
        data = self.spark.createDataFrame(
            [
                (
                    0.0,
                    5.0,
                ),
                (
                    1.0,
                    4.0,
                ),
                (
                    2.0,
                    3.0,
                ),
                (
                    2.0,
                    2.0,
                ),
                (
                    0.0,
                    1.0,
                ),
                (
                    2.0,
                    0.0,
                ),
            ],
            ["index1", "index2"],
        )
        test = self.spark.createDataFrame(
            [
                (
                    3.0,
                    7.0,
                )
            ],
            ["index1", "index2"],
        )  # invalid data

        model = encoder.fit(data)
        model_onnx = convert_sparkml(
            model,
            "Sparkml OneHotEncoder",
            [
                ("index1", FloatTensorType([None, 1])),
                ("index2", FloatTensorType([None, 1])),
            ],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)
        self.assertTrue(model_onnx.graph.node is not None)
        # run the model
        predicted = model.transform(test)
        data_np = {
            "index1": test.select("index1").toPandas().values.astype(numpy.float32),
            "index2": test.select("index2").toPandas().values.astype(numpy.float32),
        }

        predicted_np_1 = (
            predicted.select("index1Vec")
            .toPandas()
            .index1Vec.apply(lambda x: x.toArray())
            .values
        )
        predicted_np_2 = (
            predicted.select("index2Vec")
            .toPandas()
            .index2Vec.apply(lambda x: x.toArray())
            .values
        )
        expected = {
            "index1Vec": numpy.asarray(predicted_np_1.tolist()),
            "index2Vec": numpy.asarray(predicted_np_2.tolist()),
        }

        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlOneHotEncoder"
        )
        onnx_model_path = paths[-1]

        output_names = ["index1Vec", "index2Vec"]
        output, output_shapes = run_onnx_model(output_names, data_np, onnx_model_path)
        actual_output = dict(zip(output_names, output))

        compare_results(expected["index1Vec"], actual_output["index1Vec"], decimal=5)
        compare_results(expected["index2Vec"], actual_output["index2Vec"], decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from onnxconverter_common.tree_ensemble import *  # noqa


def _process_process_tree_attributes(attrs):
    # Spark may store attributes as range and not necessary list.
    # ONNX does not support this type of attribute value.
    update = {}
    wrong_types = []
    for k, v in attrs.items():
        if isinstance(v, (str, int, float, np.ndarray)):
            continue
        if isinstance(v, range):
            v = update[k] = list(v)
        if isinstance(v, list):
            if k in ("nodes_values", "nodes_hitrates", "nodes_featureids"):
                if any(map(lambda s: not isinstance(s, (float, int)), v)):
                    v = [x if isinstance(x, (float, int)) else 0 for x in v]
                    update[k] = v
            continue
        wrong_types.append(f"Unexpected type {type(v)} for attribute {k!r}.")
    if len(wrong_types) > 0:
        raise TypeError(
            "Unexpected type for one or several attributes:\n" + "\n".join(wrong_types)
        )
    if update:
        attrs.update(update)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter

_LINK_FUNCTION_TO_POST_TRANSFORM = {
    "identity": "NONE",
    "logit": "LOGISTIC",
    "ologit": "LOGISTIC",
}


def _get_post_transform(params):
    link_function = params["link_function"]
    family = params["family"]
    if family == "multinomial":
        return "SOFTMAX"
    elif link_function not in _LINK_FUNCTION_TO_POST_TRANSFORM.keys():
        raise ValueError("Link function %s not supported." % link_function)
    else:
        return _LINK_FUNCTION_TO_POST_TRANSFORM[link_function]


def _get_default_tree_attribute_pairs(is_classifier, params):
    attrs = {"post_transform": _get_post_transform(params)}
    nclasses = params["nclasses"]
    if is_classifier:
        predicted_classes = nclasses if nclasses > 2 else 1
        attrs["base_values"] = [
            params["base_score"] for _ in range(0, predicted_classes)
        ]
    else:
        attrs["n_targets"] = 1
        attrs["base_values"] = [params["base_score"]]
    for k in {
        "nodes_treeids",
        "nodes_nodeids",
        "nodes_featureids",
        "nodes_modes",
        "nodes_values",
        "nodes_truenodeids",
        "nodes_falsenodeids",
        "nodes_missing_value_tracks_true",
    }:
        attrs[k] = []
    node_attr_prefix = _node_attr_prefix(is_classifier)
    for k in {"_treeids", "_nodeids", "_ids", "_weights"}:
        attrs[node_attr_prefix + k] = []
    return attrs


def _add_node(
    attr_pairs,
    is_classifier,
    tree_id,
    node_id,
    feature_id,
    mode,
    value,
    true_child_id,
    false_child_id,
    weights,
    missing,
):
    attr_pairs["nodes_treeids"].append(tree_id)
    attr_pairs["nodes_nodeids"].append(node_id)
    attr_pairs["nodes_featureids"].append(feature_id)
    attr_pairs["nodes_modes"].append(mode)
    attr_pairs["nodes_values"].append(float(value))
    attr_pairs["nodes_truenodeids"].append(true_child_id)
    attr_pairs["nodes_falsenodeids"].append(false_child_id)
    attr_pairs["nodes_missing_value_tracks_true"].append(missing)
    if mode == "LEAF":
        node_attr_prefix = _node_attr_prefix(is_classifier)
        for i, w in enumerate(weights):
            attr_pairs[node_attr_prefix + "_treeids"].append(tree_id)
            attr_pairs[node_attr_prefix + "_nodeids"].append(node_id)
            attr_pairs[node_attr_prefix + "_ids"].append(i)
            attr_pairs[node_attr_prefix + "_weights"].append(float(w))


def _node_attr_prefix(is_classifier):
    return "class" if is_classifier else "target"


def _fill_node_attributes(tree_id, node, attr_pairs, is_classifier):
    if "leftChild" in node:
        if node["isCategorical"]:
            raise ValueError("categorical splits not supported, use one_hot_explicit")
        else:
            operator = "BRANCH_GTE"
            value = node["splitValue"]
        _add_node(
            attr_pairs=attr_pairs,
            is_classifier=is_classifier,
            tree_id=tree_id,
            mode=operator,
            value=value,
            node_id=node["id"],
            feature_id=node["colId"],
            true_child_id=node["rightChild"]["id"],
            false_child_id=node["leftChild"]["id"],
            weights=None,
            missing=(0 if node["leftward"] else 1),
        )
        _fill_node_attributes(tree_id, node["leftChild"], attr_pairs, is_classifier)
        _fill_node_attributes(tree_id, node["rightChild"], attr_pairs, is_classifier)
    else:  # leaf
        weights = [node["predValue"]]
        _add_node(
            attr_pairs=attr_pairs,
            is_classifier=is_classifier,
            tree_id=tree_id,
            value=0.0,
            node_id=node["id"],
            feature_id=0,
            mode="LEAF",
            true_child_id=0,
            false_child_id=0,
            weights=weights,
            missing=False,
        )


def assign_node_ids(node, next_id):
    if node is None:
        return next_id
    node["id"] = next_id
    next_id += 1
    next_id = assign_node_ids(node.get("leftChild", None), next_id)
    return assign_node_ids(node.get("rightChild", None), next_id)


def fill_tree_attributes(model, attr_pairs, node_attr_prefix):
    for tree in model["trees"]:
        assign_node_ids(tree["root"], 0)
        _fill_node_attributes(tree["index"], tree["root"], attr_pairs, node_attr_prefix)


def convert_regression(scope, operator, container, params):
    model = operator.raw_operator

    attr_pairs = _get_default_tree_attribute_pairs(False, params)
    fill_tree_attributes(model, attr_pairs, False)

    # add nodes
    container.add_node(
        "TreeEnsembleRegressor",
        operator.input_full_names,
        operator.output_full_names,
        op_domain="ai.onnx.ml",
        name=scope.get_unique_operator_name("TreeEnsembleRegressor"),
        **attr_pairs
    )


def convert_classifier(scope, operator, container, params):
    if params["family"] == "multinomial" and params["nclasses"] == 2:
        raise ValueError(
            "Multinomial distribution with two classes "
            "not supported, use binomial distribution."
        )
    model = operator.raw_operator

    attr_pairs = _get_default_tree_attribute_pairs(True, params)
    fill_tree_attributes(model, attr_pairs, True)

    n_trees_in_group = params["n_trees_in_group"]
    attr_pairs["class_ids"] = [
        v % n_trees_in_group for v in attr_pairs["class_treeids"]
    ]
    attr_pairs["classlabels_strings"] = params["class_labels"]

    container.add_node(
        "TreeEnsembleClassifier",
        operator.input_full_names,
        operator.output_full_names,
        op_domain="ai.onnx.ml",
        name=scope.get_unique_operator_name("TreeEnsembleClassifier"),
        **attr_pairs
    )


def convert_h2o(scope, operator, container):
    params = operator.raw_operator["params"]
    is_classifier = params["classifier"]
    if is_classifier:
        convert_classifier(scope, operator, container, params)
    else:
        convert_regression(scope, operator, container, params)


register_converter("H2OTreeMojo", convert_h2o)



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_transpose, apply_reshape
from ....common._registration import register_converter


def convert_reshape(scope, operator, container):
    from coremltools.proto.NeuralNetwork_pb2 import ReshapeLayerParams as Params

    params = operator.raw_operator.reshape

    if params.mode == Params.CHANNEL_LAST:
        intra_variable_name = scope.get_unique_variable_name(
            operator.inputs[0].full_name + "_transpose"
        )
        apply_transpose(
            scope,
            operator.inputs[0].full_name,
            intra_variable_name,
            container,
            perm=[0, 2, 3, 1],
        )
    else:
        intra_variable_name = operator.inputs[0].full_name

    N = operator.inputs[0].type.shape[0]
    if N == "None":
        N = -1
    if len(params.targetShape) == 4:
        output_shape = [int(d) for d in params.targetShape]
        output_shape[0] = N  # Overwrite bad default CoreML setting
    elif len(params.targetShape) == 3:
        output_shape = [N] + [int(d) for d in params.targetShape]
    else:
        raise ValueError(
            "The targeted shape of Reshape (name: %s) "
            "must be 3-element or 4-element array but got %s"
            % (operator.full_name, params.targetShape)
        )

    apply_reshape(
        scope=scope,
        input_name=intra_variable_name,
        output_name=operator.outputs[0].full_name,
        container=container,
        operator_name=operator.full_name,
        desired_shape=output_shape,
    )


register_converter("reshape", convert_reshape)



# SPDX-License-Identifier: Apache-2.0

import unittest
import sys
import inspect
import os
import numpy
import pandas
from pyspark.ml import Pipeline
from pyspark.ml.classification import LogisticRegression
from pyspark.ml.feature import StringIndexer, OneHotEncoder, VectorAssembler
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import StringTensorType
from tests.sparkml.sparkml_test_utils import (
    save_data_models,
    run_onnx_model,
    compare_results,
)
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlPipeline(SparkMlTestCase):
    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_model_pipeline_4_stage(self):
        this_script_dir = os.path.dirname(
            os.path.abspath(inspect.getfile(inspect.currentframe()))
        )
        input_path = os.path.join(
            this_script_dir, "data", "AdultCensusIncomeOriginal.csv"
        )
        full_data = (
            self.spark.read.format("csv")
            .options(header="true", inferschema="true")
            .load(input_path)
        )
        cols = ["workclass", "education", "marital_status"]
        training_data, test_data = (
            full_data.select("income", *cols)
            .limit(1000)
            .randomSplit([0.9, 0.1], seed=1)
        )

        stages = []
        for col in cols:
            stages.append(
                StringIndexer(
                    inputCol=col, outputCol=col + "_index", handleInvalid="skip"
                )
            )
            stages.append(
                OneHotEncoder(
                    inputCols=[col + "_index"],
                    outputCols=[col + "_vec"],
                    dropLast=False,
                )
            )

        stages.append(
            VectorAssembler(inputCols=[c + "_vec" for c in cols], outputCol="features")
        )
        stages.append(
            StringIndexer(inputCol="income", outputCol="label", handleInvalid="skip")
        )
        stages.append(LogisticRegression(maxIter=100, tol=0.0001))
        pipeline = Pipeline(stages=stages)

        model = pipeline.fit(training_data)
        model_onnx = convert_sparkml(
            model,
            "Sparkml Pipeline",
            [
                ("income", StringTensorType([None, 1])),
                ("workclass", StringTensorType([None, 1])),
                ("education", StringTensorType([None, 1])),
                ("marital_status", StringTensorType([None, 1])),
            ],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)
        self.assertTrue(model_onnx.graph.node is not None)
        # run the model
        predicted = model.transform(test_data)
        data_np = {
            "income": test_data.select("income").toPandas().values,
            "workclass": test_data.select("workclass").toPandas().values,
            "education": test_data.select("education").toPandas().values,
            "marital_status": test_data.select("marital_status").toPandas().values,
        }
        expected = [
            predicted.toPandas().label.values.astype(numpy.float32),
            predicted.toPandas().prediction.values.astype(numpy.float32),
            predicted.toPandas()
            .probability.apply(lambda x: pandas.Series(x.toArray()))
            .values.astype(numpy.float32),
        ]
        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlPipeline_4Stage"
        )
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(
            ["label", "prediction", "probability"], data_np, onnx_model_path
        )
        compare_results(expected, output, decimal=5)

    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_model_pipeline_3_stage(self):
        this_script_dir = os.path.dirname(
            os.path.abspath(inspect.getfile(inspect.currentframe()))
        )
        input_path = os.path.join(
            this_script_dir, "data", "AdultCensusIncomeOriginal.csv"
        )
        full_data = (
            self.spark.read.format("csv")
            .options(header="true", inferschema="true")
            .load(input_path)
        )
        cols = ["workclass", "education", "marital_status"]
        training_data, test_data = (
            full_data.select(*cols).limit(1000).randomSplit([0.9, 0.1], seed=1)
        )

        stages = []
        for col in cols:
            stages.append(
                StringIndexer(
                    inputCol=col, outputCol=col + "_index", handleInvalid="skip"
                )
            )
            # we need the dropLast option otherwise when assembled together (below)
            # we won't be able to expand the features without difficulties
            stages.append(
                OneHotEncoder(
                    inputCols=[col + "_index"],
                    outputCols=[col + "_vec"],
                    dropLast=False,
                )
            )

        stages.append(
            VectorAssembler(inputCols=[c + "_vec" for c in cols], outputCol="features")
        )
        pipeline = Pipeline(stages=stages)

        model = pipeline.fit(training_data)
        model_onnx = convert_sparkml(
            model,
            "Sparkml Pipeline",
            [
                ("workclass", StringTensorType([None, 1])),
                ("education", StringTensorType([None, 1])),
                ("marital_status", StringTensorType([None, 1])),
            ],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)
        self.assertTrue(model_onnx.graph.node is not None)
        # run the model
        predicted = model.transform(test_data)
        data_np = {
            "workclass": test_data.select("workclass").toPandas().values,
            "education": test_data.select("education").toPandas().values,
            "marital_status": test_data.select("marital_status").toPandas().values,
        }
        expected = (
            predicted.toPandas()
            .features.apply(lambda x: pandas.Series(x.toArray()))
            .values
        )
        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlPipeline_3Stage"
        )
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(["features"], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)

    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_model_pipeline_2_stage(self):
        this_script_dir = os.path.dirname(
            os.path.abspath(inspect.getfile(inspect.currentframe()))
        )
        input_path = os.path.join(
            this_script_dir, "data", "AdultCensusIncomeOriginal.csv"
        )
        full_data = (
            self.spark.read.format("csv")
            .options(header="true", inferschema="true")
            .load(input_path)
        )
        cols = ["workclass", "education", "marital_status"]
        training_data, test_data = (
            full_data.select(*cols).limit(1000).randomSplit([0.9, 0.1], seed=1)
        )

        stages = []
        for col in cols:
            stages.append(
                StringIndexer(
                    inputCol=col, outputCol=col + "_index", handleInvalid="skip"
                )
            )
            stages.append(
                OneHotEncoder(
                    inputCols=[col + "_index"],
                    outputCols=[col + "_vec"],
                    dropLast=False,
                )
            )

        pipeline = Pipeline(stages=stages)

        model = pipeline.fit(training_data)
        model_onnx = convert_sparkml(
            model,
            "Sparkml Pipeline",
            [
                ("workclass", StringTensorType([None, 1])),
                ("education", StringTensorType([None, 1])),
                ("marital_status", StringTensorType([None, 1])),
            ],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)
        self.assertTrue(model_onnx.graph.node is not None)
        # run the model
        predicted = model.transform(test_data)
        data_np = {
            "workclass": test_data.select("workclass").toPandas().values,
            "education": test_data.select("education").toPandas().values,
            "marital_status": test_data.select("marital_status").toPandas().values,
        }
        predicted_np = [
            predicted.toPandas()
            .workclass_vec.apply(lambda x: pandas.Series(x.toArray()))
            .values,
            predicted.toPandas()
            .education_vec.apply(lambda x: pandas.Series(x.toArray()))
            .values,
            predicted.toPandas()
            .marital_status_vec.apply(lambda x: pandas.Series(x.toArray()))
            .values,
        ]

        expected = [numpy.asarray(row) for row in predicted_np]
        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlPipeline_2Stage"
        )
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(
            ["workclass_vec", "education_vec", "marital_status_vec"],
            data_np,
            onnx_model_path,
        )
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

from uuid import uuid4
import onnx
from ..common.onnx_ex import get_maximum_opset_supported
from ..common._topology import convert_topology
from ._parse import parse_sparkml
from . import operator_converters  # noqa: F401


def convert(
    model,
    name=None,
    initial_types=None,
    doc_string="",
    target_opset=None,
    targeted_onnx=onnx.__version__,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
    spark_session=None,
):
    """
    This function produces an equivalent ONNX model
    of the given spark-ml model. The supported spark-ml
    modules are listed below.

    * Preprocessings and transformations:
      1.  pyspark.ml.feature.DictVectorizer
      3.  preprocessing.LabelEncoder
      5.  preprocessing.OneHotEncoder

    * Linear classification and regression:
      9.  pyspark.ml.classification.LinearRegression

    * Support vector machine for classification and regression

    * Tree-based models for classification and regression

    * pipeline
      29. pipeline.Pipeline

    For pipeline conversion, user needs to make sure
    each component is one of our supported items (1)-(24).

    This function converts the specified spark-ml model into its
    ONNX counterpart. Notice that for all conversions,
    initial types are required.  ONNX model name can also be specified.

    :param model: A spark-ml model
    :param initial_types: a python list. Each element is a
        tuple of a variable name and a type defined in data_types.py
    :param name: The name of the graph (type: GraphProto)
        in the produced ONNX model (type: ModelProto)
    :param doc_string: A string attached onto the produced ONNX model
    :param target_opset: number, for example, 7 for ONNX 1.2, and 8 for ONNX 1.3.
    :param targeted_onnx: A string (for example, '1.1.2' and '1.2')
        used to specify the targeted ONNX version of the
        produced model. If ONNXMLTools cannot find a compatible
        ONNX python package, an error may be thrown.
    :param custom_conversion_functions: a dictionary for
        specifying the user customized conversion function
    :param custom_shape_calculators: a dictionary for
        specifying the user customized shape calculator
    :return: An ONNX model (type: ModelProto) which is equivalent to the input spark-ml model

    Example of initial_types:

    Assume that the specified spark-ml model takes a heterogeneous
    list as its input. If the first 5 elements are
    floats and the last 10 elements are integers, we need to
    specify initial types as below. The [1] in [1, 5] indicates
    the batch size here is 1.
    >>> from onnxmltools.convert.common.data_types import FloatTensorType, Int64TensorType
    >>> initial_type = [('float_input', FloatTensorType([1, 5])),
        ('int64_input', Int64TensorType([1, 10]))]
    """
    if initial_types is None:
        raise ValueError(
            "Initial types are required. See usage of convert(...) in \
                         onnxmltools.convert.sparkml.convert for details"
        )

    if name is None:
        name = str(uuid4().hex)

    target_opset = target_opset if target_opset else get_maximum_opset_supported()
    # Parse spark-ml model as our internal data structure (i.e., Topology)
    topology = parse_sparkml(
        spark_session,
        model,
        initial_types,
        target_opset,
        custom_conversion_functions,
        custom_shape_calculators,
    )

    # Infer variable shapes
    topology.compile()

    # Convert our Topology object into ONNX. The outcome is an ONNX model.
    onnx_model = convert_topology(
        topology, name, doc_string, target_opset, targeted_onnx
    )

    return onnx_model



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from .Scale import deduce_broadcast_axis_and_shape
from .....proto import onnx_proto
from ....common._apply_operation import apply_add, apply_reshape, apply_cast
from ....common._registration import register_converter


def convert_embedding(scope, operator, container):
    params = operator.raw_operator.embedding
    gather_op_name = scope.get_unique_operator_name("Gather")
    gather_attrs = {"name": gather_op_name}

    # Reshape the indexes we want to embed to 1-D tensor.
    # Otherwise, ONNX Gather's output may get wrong shape.
    reshaped_input_name = scope.get_unique_variable_name(
        gather_op_name + "input_reshaped"
    )  # 2nd input of Gather
    apply_reshape(
        scope,
        operator.inputs[0].full_name,
        reshaped_input_name,
        container,
        desired_shape=[-1],
    )

    # ONNX Gather accepts integers so we add a Cast to enforce
    # this before feeding input into ONNX Gather.
    casted_input_name = scope.get_unique_variable_name(
        gather_op_name + "input_casted"
    )  # 2nd input of Gather
    apply_cast(
        scope,
        reshaped_input_name,
        casted_input_name,
        container,
        to=onnx_proto.TensorProto.INT64,
    )

    # Load the embedding matrix. Its shape is outputChannels-by-inputDim.
    weights = np.array(params.weights.floatValue).reshape(
        params.outputChannels, params.inputDim
    )
    weights_name = scope.get_unique_variable_name(
        gather_op_name + "_W"
    )  # 1st input of Gather
    container.add_initializer(
        weights_name,
        onnx_proto.TensorProto.FLOAT,
        [params.inputDim, params.outputChannels],
        weights.transpose().flatten().tolist(),
    )

    # To support the bias term in an embedding (if exists), we need to create one extra node.
    if params.hasBias:
        # Put the embedded result onto a temporal tensor
        gather_output_name = scope.get_unique_variable_name(gather_op_name + "_output")
        container.add_node(
            "Gather",
            [weights_name, casted_input_name],
            gather_output_name,
            **gather_attrs
        )

        # Load the bias vector into an initializer
        bias_name = scope.get_unique_variable_name(gather_op_name + "_bias")
        bias_axis, bias_shape = deduce_broadcast_axis_and_shape(
            container.target_opset, [params.outputChannels]
        )
        container.add_initializer(
            bias_name, onnx_proto.TensorProto.FLOAT, bias_shape, params.bias.floatValue
        )
        # Create an addition operator to add bias (shape: [C])
        # into Gather's tensor (shape: [N, C])
        apply_add(
            scope,
            [gather_output_name, bias_name],
            operator.outputs[0].full_name,
            container,
            axis=1,
            broadcast=1,
        )
    else:
        # This case has no bias, so we just output the result produced by the embedding node.
        container.add_node(
            "Gather",
            [weights_name, casted_input_name],
            operator.output_full_names,
            **gather_attrs
        )


register_converter("embedding", convert_embedding)



# SPDX-License-Identifier: Apache-2.0

from .....proto import onnx_proto
from ....common._registration import register_converter
from ....common._apply_operation import apply_add
from .Scale import deduce_broadcast_axis_and_shape


def convert_bias(scope, operator, container):
    # Feed the input (which we are going to add a bias onto)
    # into Add operator. Its shape is [C, H, W] in CoreML but
    # [N, C, H, W] in ONNX.
    params = operator.raw_operator.bias

    # Adjust CoreML's bias shape and find a proper axis for broadcasting
    axis, shape = deduce_broadcast_axis_and_shape(container.target_opset, params.shape)

    # No matter what shape it is, we need "broadcast" on
    # because input shape is 4-D while bias is at most 3-D.
    broadcast = 1  # True

    # Create bias vector as an ONNX tensor
    bias_tensor_name = scope.get_unique_variable_name(operator.full_name + "_B")
    container.add_initializer(
        bias_tensor_name, onnx_proto.TensorProto.FLOAT, shape, params.bias.floatValue
    )

    apply_add(
        scope,
        [operator.inputs[0].full_name, bias_tensor_name],
        operator.output_full_names,
        container,
        operator_name=operator.full_name,
        axis=axis,
        broadcast=broadcast,
    )


register_converter("bias", convert_bias)



# SPDX-License-Identifier: Apache-2.0

from onnx import onnx_pb as onnx_proto
from ...common.data_types import FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._apply_operation import apply_mul
from ...common._registration import register_converter, register_shape_calculator


def convert_element_wise_product(scope, operator, container):
    op = operator.raw_operator
    scaling_vector = scope.get_unique_variable_name("scaling_vector")
    container.add_initializer(
        scaling_vector,
        onnx_proto.TensorProto.FLOAT,
        [1, len(op.getScalingVec())],
        op.getScalingVec(),
    )
    apply_mul(
        scope,
        [operator.inputs[0].full_name, scaling_vector],
        operator.output_full_names,
        container,
    )


register_converter(
    "pyspark.ml.feature.ElementwiseProduct", convert_element_wise_product
)


def calculate_element_wise_product_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])
    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = FloatTensorType([N, operator.inputs[0].type.shape[1]])


register_shape_calculator(
    "pyspark.ml.feature.ElementwiseProduct",
    calculate_element_wise_product_output_shapes,
)



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import numpy
import pandas
from pyspark.ml.feature import ChiSqSelector
from pyspark.ml.linalg import Vectors
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import (
    save_data_models,
    run_onnx_model,
    compare_results,
)
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlChiSqSelector(SparkMlTestCase):
    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_chi_sq_selector(self):
        data = self.spark.createDataFrame(
            [
                (Vectors.dense([0.0, 0.0, 18.0, 1.0]), 1.0),
                (Vectors.dense([0.0, 1.0, 12.0, 0.0]), 0.0),
                (Vectors.dense([1.0, 0.0, 15.0, 0.1]), 0.0),
            ],
            ["features", "label"],
        )
        selector = ChiSqSelector(numTopFeatures=1, outputCol="selectedFeatures")
        model = selector.fit(data)

        # the input name should match that of what StringIndexer.inputCol
        feature_count = data.first()[0].size
        model_onnx = convert_sparkml(
            model,
            "Sparkml ChiSqSelector",
            [("features", FloatTensorType([None, feature_count]))],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)

        # run the model
        predicted = model.transform(data)
        expected = (
            predicted.toPandas()
            .selectedFeatures.apply(lambda x: pandas.Series(x.toArray()))
            .values.astype(numpy.float32)
        )
        data_np = (
            data.toPandas()
            .features.apply(lambda x: pandas.Series(x.toArray()))
            .values.astype(numpy.float32)
        )
        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlChiSqSelector"
        )
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(
            ["selectedFeatures"], data_np, onnx_model_path
        )
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

from ....proto import onnx_proto
from ...common._registration import register_converter
from ...common.utils import cast_list
import numpy

try:
    from libsvm import svm
except ImportError:
    # Older version of libsvm.
    import svm


class SVMConverter:
    """
    Converts a SVM model trained with *svmlib*.
    """

    @staticmethod
    def validate(svm_node):
        try:
            hasattr(svm_node, "param")
            hasattr(svm_node, "SV")
            hasattr(svm_node, "nSV")
            hasattr(svm_node, "sv_coef")
            hasattr(svm_node, "l")
            hasattr(svm_node.param, "gamma")
            hasattr(svm_node.param, "coef0")
            hasattr(svm_node.param, "degree")
            hasattr(svm_node.param, "kernel_type")
            hasattr(svm_node, "rho")
        except AttributeError as e:
            raise RuntimeError("Missing type from svm node:" + str(e))

    @staticmethod
    def get_sv(svm_node):
        labels = svm_node.get_labels()
        sv = svm_node.get_SV()
        if len(sv) == 0:
            raise RuntimeError(
                "No support vector machine. This usually happens "
                "with very small datasets or the training failed."
            )

        maxk = max(max(row.keys() for row in sv))
        mat = numpy.zeros((len(sv), maxk + 1), dtype=numpy.float32)

        for i, row in enumerate(sv):
            for k, v in row.items():
                if k == -1:
                    k = 0
                try:
                    mat[i, k] = v
                except IndexError:
                    raise RuntimeError(
                        "Issue with one dimension\nlabels={0}\n#sv={1}\n"
                        "shape={2}\npos={3}x{4}-maxk={5}-svm.l={6}\nrow={7}".format(
                            labels, sv, mat.shape, i, k, maxk, svm_node.l, row
                        )
                    )
        # We do not consider the first row (class -1).
        mat = mat[:, 1:]

        # mat.shape should be (n_vectors, X.shape[1])
        # where X.shape[1] is the number of features.
        # However, it can be <= X.shape.[1] if the last
        # every coefficient is null on the last column.
        # To fix that, an extra parameter must be added to
        # the convert function as there is no way to guess
        # that information from svmlib model.
        return numpy.array(mat.ravel(), dtype=float)

    @staticmethod
    def convert(operator, scope, container, svm_node, inputs, model_name, nb_class):
        kt = svm_node.param.kernel_type
        if kt == svm.RBF:
            kt = "RBF"
        elif kt == svm.SIGMOID:
            kt = "SIGMOID"
        elif kt == svm.POLY:
            kt = "POLY"
        elif kt == svm.LINEAR:
            kt = "LINEAR"
        else:
            raise RuntimeError("Unexpected value for kernel: {0}".format(kt))

        def copy_sv_coef(sv_coef):
            nrc = svm_node.nr_class - 1
            res = numpy.zeros((svm_node.l, nrc), dtype=numpy.float64)
            for i in range(0, svm_node.l):
                for j in range(nrc):
                    res[i, j] = svm_node.sv_coef[j][i]
            return res.T

        if nb_class > 2:
            # See above.
            coef = copy_sv_coef(svm_node.sv_coef)
        else:
            coef = numpy.array(svm_node.get_sv_coef()).ravel()

        atts = dict(
            kernel_type=kt,
            kernel_params=[
                float(_)
                for _ in [
                    svm_node.param.gamma,
                    svm_node.param.coef0,
                    svm_node.param.degree,
                ]
            ],
            coefficients=list(coef.ravel()),
        )

        return dict(
            node="SVMConverter",
            inputs=operator.input_full_names,
            outputs=[o.full_name for o in operator.outputs],
            op_domain="ai.onnx.ml",
            attrs=atts,
        )


class SVCConverter(SVMConverter):
    @staticmethod
    def validate(svm_node):
        SVMConverter.validate(svm_node)
        try:
            hasattr(svm_node, "probA")
            hasattr(svm_node, "probB")
        except AttributeError as e:
            raise RuntimeError("Missing type from svm node:" + str(e))

    @staticmethod
    def convert(operator, scope, container, svm_node, inputs):
        nbclass = len(svm_node.get_labels())
        # See converter for sklearn.
        nb = SVMConverter.convert(
            operator, scope, container, svm_node, inputs, "SVMClassifier", nbclass
        )
        sign_rho = -1.0
        st = svm_node.param.svm_type

        if svm_node.is_probability_model():
            if st == svm.C_SVC or st == svm.NU_SVC:
                n_class = len(svm_node.get_labels())
                n = int(n_class * (n_class - 1) / 2)
                probA = [svm_node.probA[i] for i in range(n)]
                probB = [svm_node.probB[i] for i in range(n)]
                nb["attrs"]["prob_a"] = probA
                nb["attrs"]["prob_b"] = probB
                nb["attrs"]["rho"] = [svm_node.rho[i] * sign_rho for i in range(n)]
            else:
                nb["attrs"]["rho"] = [svm_node.rho[0] * sign_rho]
        elif st == svm.C_SVC or st == svm.NU_SVC:
            n_class = len(svm_node.get_labels())
            n = int(n_class * (n_class - 1) / 2)
            nb["attrs"]["rho"] = [svm_node.rho[i] * sign_rho for i in range(n)]
        else:
            nb["attrs"]["rho"] = [svm_node.rho[0] * sign_rho]

        class_labels = cast_list(int, svm_node.get_labels())
        # Predictions are different when label are not sorted (multi-classification).
        class_labels.sort()
        nb["attrs"]["classlabels_ints"] = class_labels
        onnx_proto.TensorProto.INT64

        if len(nb["outputs"]) != 2:
            raise RuntimeError(
                "The model outputs label and probabilities not {0}".format(
                    nb["outputs"]
                )
            )

        nbclass = len(svm_node.get_labels())
        nb["attrs"]["vectors_per_class"] = [svm_node.nSV[i] for i in range(nbclass)]
        nb["attrs"]["post_transform"] = "NONE"
        nb["attrs"]["support_vectors"] = SVCConverter.get_sv(svm_node)

        # Add a vec dictionizer to handle the map output
        container.add_node(
            "SVMClassifier",
            nb["inputs"],
            nb["outputs"],
            op_domain="ai.onnx.ml",
            name=scope.get_unique_operator_name("SVMClassifier"),
            **nb["attrs"]
        )


class SVRConverter(SVMConverter):
    @staticmethod
    def validate(svm_node):
        SVMConverter.validate(svm_node)
        try:
            hasattr(svm_node, "l")
        except AttributeError as e:
            raise RuntimeError("Missing type from svm node:" + str(e))

    @staticmethod
    def convert(operator, scope, container, svm_node, inputs):
        nb = SVMConverter.convert(
            operator, scope, container, svm_node, inputs, "SVMRegressor", 0
        )

        nb["attrs"]["n_supports"] = svm_node.l
        nb["attrs"]["post_transform"] = "NONE"
        nb["attrs"]["rho"] = [-svm_node.rho[0]]
        nb["attrs"]["support_vectors"] = SVCConverter.get_sv(svm_node)

        container.add_node(
            "SVMRegressor",
            nb["inputs"],
            nb["outputs"],
            op_domain="ai.onnx.ml",
            name=scope.get_unique_operator_name("SVMRegressor"),
            **nb["attrs"]
        )


class AnyLibSvmConverter:
    @staticmethod
    def select(svm_node):
        if svm_node.param.svm_type in (svm.C_SVC, svm.NU_SVC):
            return SVCConverter
        if svm_node.param.svm_type in (svm.EPSILON_SVR, svm.NU_SVR):
            return SVRConverter
        raise RuntimeError(
            "svm_node type is unexpected '{0}'".format(svm_node.param.svm_type)
        )

    @staticmethod
    def validate(svm_node):
        sel = AnyLibSvmConverter.select(svm_node)
        sel.validate(svm_node)

    @staticmethod
    def convert(operator, scope, container, svm_node, inputs):
        sel = AnyLibSvmConverter.select(svm_node)
        sel.convert(operator, scope, container, svm_node, inputs)


def convert_libsvm(scope, operator, container):
    inputs = operator.inputs
    model = operator.raw_operator
    converter = AnyLibSvmConverter
    converter.validate(model)
    converter.convert(operator, scope, container, model, inputs)


# Register the class for processing
register_converter("LibSvmSVC", convert_libsvm)
register_converter("LibSvmSVR", convert_libsvm)



# SPDX-License-Identifier: Apache-2.0

from ....common._registration import register_converter


def convert_mean_variance_normalization(scope, operator, container):
    op_type = "MeanVarianceNormalization"
    op_name = scope.get_unique_operator_name(op_type)
    params = operator.raw_operator.mvn
    attrs = {"name": op_name}
    attrs["across_channels"] = params.acrossChannels
    attrs["normalize_variance"] = params.normalizeVariance
    container.add_node(
        op_type, operator.input_full_names, operator.output_full_names, **attrs
    )


register_converter("mvn", convert_mean_variance_normalization)
